################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
IndexedFasta.py - fast random access in fasta files
===================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Usage
-----


Code
----

'''
import os, sys, array, string, re, types, optparse, time, struct, math, tarfile, logging
import platform, anydbm
import math
import random
import zlib
import gzip
import cStringIO
import Experiment as E
from AString import AString
import pysam

##------------------------------------------------------------
class Uncompressor:
    def __init__(self, filename, unmangler):
        self.mFile = open(filename, "rb" )
        self.mUnMangler = unmangler
        
    def read( self, block_size, indices, start, end ):
        """read an uncompressed block from start:end.
        
        The compressed chunk starts at first_pos.
        NOTE: This is poorly implemented - do better.
        """

        # skip over uncompressed blocks
        d = int(math.floor(float(start) / block_size) )
        r = start % block_size
        assert( d < len(indices) )
        self.mFile.seek( indices[d] )

        # read x bytes of compressed data, at least one full chunk.
        nchunks = int(math.ceil( float((r+end-start)) / block_size) )

        fragments = []
        for x in range(d, d+nchunks):
            s = self.mFile.read( indices[x+1] - indices[x] )
            fragments.append(self.mUnMangler( s ))
        u = "".join(fragments)

        assert len(u) >= end - start, "fragment smaller than requested size: %i > %i-%i=%i" % (len(u), end,start,end-start)
        
        return u[r:r+end-start]

##------------------------------------------------------------    
def writeFragments( outfile_fasta, 
                    outfile_index,
                    fragments, 
                    mangler, size,
                    write_all = False):
    """write mangled fragments to *outfile_fasta* in chunks of *size*
    updating *outfile_index*.

    returns part of last fragment that has not been written and is 
    less than *size* and the number of fragments output.

    If *write_all* is True, all of the fragments are written to
    the file and the last file position is added to *outfile_index*
    as well.
    """

    s = "".join(fragments)
    rest = len(s) % size
    if len(s) > size:
        for x in range(0, len(s)-rest, size):
            outfile_index.write( "\t%i" % outfile_fasta.tell() )
            outfile_fasta.write( mangler(s[x:x+size]) )
        
    if rest:
        if write_all:
            outfile_index.write("\t%i" % outfile_fasta.tell() )
            outfile_fasta.write( mangler( s[-rest:] ) )
            outfile_index.write("\t%i" % outfile_fasta.tell() )            
            return ""
        else:
            return s[-rest:]
    else:
        return ""

def gzip_mangler(s):

    xfile = cStringIO.StringIO()

    gzipfile = gzip.GzipFile( fileobj = xfile, mode="wb" )
    gzipfile.write( s )
    gzipfile.close()
    
    m = xfile.getvalue()
    xfile.close()
    return m

def gzip_demangler(s):

    gzipfile = gzip.GzipFile( fileobj = cStringIO.StringIO(s), mode="rb" )
    m = gzipfile.readline()
    return m

##------------------------------------------------------------
class Translator:
    """translate a sequence."""
    def __init__(self): 
        self.mRegEx = re.compile( " +" )

    def __call__(self, sequence):
        return "".join( self.mMapScore2Char[self.mMapScore2Score[int(x)]] for x in self.mRegEx.split( sequence.strip() ) ) 

    def translate( self, sequence ):
        raise NotImplementedError( "translate not implemented.")

class TranslatorPhred(Translator):
    """translate phred quality scores."""
    def __init__(self, *args, **kwargs):
        Translator.__init__(self,*args,**kwargs)
        self.mMapScore2Char = [ chr(33 + x) for x in range( 0, 93) ]
        self.mMapScore2Score = range( 0, 93 )

    def translate( self, sequence ):
        return array.array( "I", ( ord(x)-33 for x in sequence ) )

class TranslatorSolexa(Translator):
    """translate solexa quality scores."""
    def __init__(self, *args, **kwargs):
        Translator.__init__(self,*args,**kwargs)
        self.mMapScore2Char = [ chr(64 + x) for x in range( 0, 128) ]
        self.mMapScore2Score =[ int(10.0 * math.log( 1.0 + 10 ** (x / 10.0)) / math.log(10)+.499) for x in range(-64,65) ]

    def translate( self, sequence ):
        raise NotImplementedError( "translate not implemented.")
        return array.array( "i", ( ord(x) - 64 ) )

class TranslatorRange200(Translator):
    """translate pcap quality scores.

    For example for PCAP scores.

    These scores range from 0 to 100 and are the
    "a weighted sum of input base quality values
    (Huang and Madan 1999)

    The numerical values from 0 to 200 are stored
    as values form 33 to 233
    "
    """
    def __init__(self, *args, **kwargs):
        Translator.__init__(self,*args,**kwargs)
        self.mMapScore2Char = [ chr(33 + x) for x in range( 0, 200) ]

    def __call__(self, sequence):
        try:
            return "".join( self.mMapScore2Char[int(x)] for x in self.mRegEx.split( sequence.strip() ) ) 
        except ValueError, msg:
            raise ValueError( msg + " parsing error in fragment: %s" % sequence )

    def translate( self, sequence ):
        return array.array( "I", ( ord(x) - 33 for x in sequence ) )

class TranslatorBytes(Translator):
    """output binary values as bytes permitting values from 0 to 255

    Note the resulting file will not be iterable as newline is not
    a record-separator any more.
    """
    def __init__(self, *args, **kwargs):
        Translator.__init__(self,*args,**kwargs)

    def __call__(self, sequence):
        try:
            return "".join( chr(int(x)) for x in self.mRegEx.split( sequence.strip() ) ) 
        except ValueError, msg:
            print "parsing error in line: %s" % sequence
            print "message=%s" % str(msg)
            return ""
        
    def translate( self, sequence ):
        return array.array( "I", ( ord(x) for x in sequence ) )

##------------------------------------------------------------
class MultipleFastaIterator:

    def __init__(self, 
                 filenames, 
                 regex_identifier = None,
                 format = "auto" ):

        if type(filenames) == types.StringType:
            self.filenames = [filenames]
        else:
            self.filenames = filenames

        self.regexIdentifier = regex_identifier
        self.iterator = self._iterate()
        self.format = format

    def __iter__(self):
        return self

    def next(self):
        try:
            return self.iterator.next( )
        except StopIteration:
            return None

    def _iterate( self ):
        """iterate over muliple files."""
        
        def _iter( infile ):

            identifier = None

            for line in infile:
                if line.startswith("#"):  continue
                if line.startswith(">"):

                    if self.regexIdentifier:
                        try:
                            identifier = re.search(self.regexIdentifier, line[1:-1]).groups()[0]
                        except AttributeError:
                            raise ValueError("could not parse identifier from line %s - check the input" % line[1:-1])
                    else:
                        identifier = re.split("\s", line[1:-1])[0]

                else:
                    if not identifier:
                        raise ValueError("refusing to emit sequence without identifier - check the input")
                    yield identifier, line.strip()

        for filename in self.filenames:
            if self.format == "tar.gz" or self.format == "tar" or (self.format == "auto" and filename.endswith( "tar.gz" )):
                if filename == "-":
                    tf = tarfile.open( fileobj = sys.stdin, mode = "r|*" ) 
                else:
                    tf = tarfile.open( filename, mode = "r" )
                for f in tf:
                    b, ext = os.path.splitext( f.name )
                    if ext.lower() in ( ".fasta", ".fa" ):
                        E.info( "extracting %s" % f.name)
                        infile = tf.extractfile( f )
                        for x in _iter( infile ): yield x
                    else:
                        E.info( "skipping %s" % f.name )

                if tf != sys.stdin: tf.close()
                continue
            elif self.format == "fasta.gz" or (self.format == "auto" and filename.endswith(".gz")):
                infile = gzip.open( filename, "r" )
            elif filename == "-":
                infile = sys.stdin
            else:
                infile = open( filename, "r")

            for x in _iter( infile ): yield x
            if filename != "-": infile.close()

        raise StopIteration

##------------------------------------------------------------
def createDatabase( db, iterator,
                    force = False,
                    synonyms = None,
                    compression = None,
                    random_access_points = None,
                    regex_identifier = None,
                    clean_sequence = False,
                    ignore_duplicates = False,
                    allow_duplicates = False,
                    translator = None ):
    """index files in filenames to create database.

    Two new files are created - db.fasta and db_name.idx

    If compression is enabled, provide random access points
    every # bytes.

    Dictzip is treated as an uncompressed file.

    regex_identifier: pattern to extract identifier from description line.
    If None, the part until the first white-space character is used.

    translator: specify a translator
    """

    if db.endswith( ".fasta"): db = db[:-len(".fasta")]

    if compression:
        if compression == "lzo":
            import lzo
            def lzo_mangler( s ): return lzo.compress(s, 9)
            mangler = lzo_mangler
            db_name = db + ".lzo"
            write_chunks = True
        elif compression == "zlib":
            def zlib_mangler( s ): return zlib.compress( s, 9)
            mangler = zlib_mangler
            db_name = db + ".zlib"
            write_chunks = True     
        elif compression == "gzip":
            mangler = gzip_mangler
            db_name = db + ".gz"
            write_chunks = True     
        elif compression == "dictzip":
            import dictzip
            mangler = lambda x: x
            db_name = db + ".dz"
            write_chunks = False
        elif compression == "bzip2":
            import bz2
            def bzip_mangler( s ): return bz2.compress( s, 9)
            mangler = bzip_mangler
            db_name = db + ".bz2"
            write_chunks = True
        elif compression == "debug":
            mangler = lambda x: x
            db_name = db + ".debug"
            write_chunks = True
        elif compression == "rle":
            import RLE
            mangler = RLE.compress
            db_name = db + ".rle"
            write_chunks = True
        else:
            raise ValueError("unknown compression library: %s" % compression)

        index_name = db + ".cdx"

        if write_chunks and random_access_points == None or random_access_points <= 0:
            raise ValueError("specify chunksize in --random-access-points")
        
    else:
        mangler = lambda x: x
        db_name = db + ".fasta"
        write_chunks = False
        index_name = db + ".idx"
    
    if os.path.exists( db_name ) and not force:
        raise ValueError( "database %s already exists." % db_name )

    if os.path.exists( index_name ) and not force:
        raise ValueError( "database index %s already exists." % index_name )
    
    outfile_index = open( index_name, "w" )
    if compression == "dictzip":
        import dictzip
        if random_access_points == None or random_access_points <= 0:
            raise ValueError("specify dictzip chunksize in --random-access-points")
        outfile_fasta = dictzip.open( db_name, "wb", buffersize=1000000, chunksize=random_access_points )
        compression = None
    else:
        outfile_fasta = open( db_name, "wb" )

    identifiers = {}
    lsequence = 0
    identifier_pos, sequence_pos = 0, 0

    translation = string.maketrans("xX", "nN")

    fragments = []
    lfragment = 0
    last_identifier = None

    while 1:

        try:
            result = iterator.next()
        except StopIteration: 
            break

        if not result: break

        identifier, fragment = result

        if identifier != last_identifier:

            ## check for duplicate identifiers
            if identifier in identifiers:
                if ignore_duplicates:
                    raise ValueError, "ignore duplicates not implemented"
                elif allow_duplicates:
                    # the current implementation will fail if the same identifiers
                    # are directly succeeding each other
                    # better: add return to iterator that indicates a new identifier
                    out_identifier = identifier + "_%i" % (identifiers[identifier])
                    identifiers[identifier] += 1
                    identifiers[out_identifier] = 1
                else:
                    raise ValueError, "%s occurs more than once" %\
                        (identifier,)
            else:
                identifiers[identifier] = 1
                out_identifier = identifier

            if last_identifier:
                if write_chunks:
                    writeFragments( outfile_fasta, outfile_index, 
                                    fragments, mangler,
                                    size = random_access_points, 
                                    write_all = True )
                    
                    fragments = []
                    lfragment = 0
                else:
                    outfile_fasta.write( "\n" )
                        
                outfile_index.write("\t%i\n" % lsequence)

            # write identifier
            identifier_pos = outfile_fasta.tell()
            outfile_fasta.write( mangler(">%s\n" % out_identifier) )
            sequence_pos = outfile_fasta.tell()
                
            outfile_index.write( "%s\t%i" % (out_identifier,
                                             identifier_pos ) )
            if write_chunks:
                outfile_index.write( "\t%i" % random_access_points )
            else:
                outfile_index.write( "\t%i" % sequence_pos )

            fragments = []
            lsequence = 0
            last_identifier = identifier
                
        if translator:
            s = translator( fragment )
        else:
            s = re.sub( "\s", "", fragment.strip() )
            if clean_sequence:
                s = s.translate( translation )
                        
        lsequence += len(s)
                
        if write_chunks:
            fragments.append(s)
            lfragment += len(s)
            if lfragment > random_access_points:
                rest = writeFragments( outfile_fasta, 
                                       outfile_index,
                                       fragments, 
                                       mangler, 
                                       size = random_access_points,
                                       write_all = False)
                fragments = [rest]
                lfragment = len(rest)
        else:
            outfile_fasta.write( mangler(s) )
                    
    if write_chunks:
        writeFragments( outfile_fasta, outfile_index, fragments, mangler, 
                        size = random_access_points, write_all = True )
    else:
        outfile_fasta.write( "\n" )
            
    outfile_index.write("\t%i\n" % lsequence )

    # add synonyms for the table
    if synonyms:
        for key, vals in synonyms.items():
            for val in vals:
                outfile_index.write( "%s\t%s\n" % (key, val) )

# map of names
# order is suffix data, suffix index, noSeek
NAME_MAP={
    'uncompressed' : ('fasta', 'idx', False),
    'lzo'          : ('lzo',   'cdx', True ),    
    'dictzip'      : ('dz',    'idx', False ),
    'zlib'         : ('zlib',  'cdx', True ),
    'gzip'         : ('gzip',  'cdx', True ),
    'bzip2'        : ('bz2',   'cdx', True ),
    'debug'        : ('debug', 'cdx', True ),
    }

PREFERENCES=('uncompressed', 'lzo', 'dictzip', 'zlib', 'gzip', 'bzip2', 'debug')

class CGATIndexedFasta:
    """an indexed fasta file."""

    def __init__( self, dbname ):

        if dbname.endswith( ".fasta"): dbname = dbname[:-len(".fasta")]

        for x in PREFERENCES:
            d =  "%s.%s" % (dbname, NAME_MAP[x][0] )
            i =  "%s.%s" % (dbname, NAME_MAP[x][1] )
            if os.path.exists( d ) and os.path.exists( i ):
                self.mMethod = x
                self.mDbname = d
                self.mNameIndex = i
                self.mNoSeek = NAME_MAP[x][2]
                break
        else:
            raise KeyError, "unknown database %s" % dbname 
        
        self.mIsLoaded = False
        self.mSynonyms = {} 
        self.mConverter = None
        self.mIndex = {}
        self.mTranslator = None

    def __len__(self):
        """return the number of sequences in fasta file."""
        if not self.mIsLoaded: self._loadIndex()
        return len(self.mIndex)

    def __contains__(self, contig):
        if not self.mIsLoaded: self._loadIndex()
        return contig in self.mIndex or contig in self.mSynonyms

    def __getitem__(self, key ):
        """return full length sequence."""
        return self.getSequence( key, "+", 0, 0, as_array = True )
        
    def _loadIndex( self, compress=False ):
        """load complete index into memory.

        if compress is set to true, the index will not be loaded,
        but a compressed index will be created instead.
        """

        if self.mMethod == "uncompressed":
            self.mDatabaseFile = open( self.mDbname, "r" )
        elif self.mMethod == "dictzip":
            import dictzip
            self.mDatabaseFile = dictzip.GzipFile( self.mDbname)
        elif self.mMethod == "lzo":
            import lzo
            self.mDatabaseFile = Uncompressor( self.mDbname, lzo.decompress )
        elif self.mMethod == "gzip":
            self.mDatabaseFile = Uncompressor( self.mDbname, gzip_demangler )
        elif self.mMethod == "zlib":
            self.mDatabaseFile = Uncompressor( self.mDbname, zlib.decompress )
        elif self.mMethod == "bzip2":
            import bz2
            self.mDatabaseFile = Uncompressor( self.mDbname, bz2.decompress )
        elif self.mMethod == "debug":
            self.mDatabaseFile = Uncompressor( self.mDbname + ".debug", lambda x: x )            

        filename_index = self.mNameIndex + ".dbm"
        
        if compress:
            if os.path.exists( filename_index ):
                raise OSError( "file %s already exists" % filename_index )
            self.mIndex = anydbm.open( filename_index, "n" )
        elif os.path.exists( filename_index ):
            self.mIndex = anydbm.open( filename_index, "r" )
            self.mIsLoaded = True
            return
        else:
            self.mIndex = {}

        for line in open(self.mNameIndex, "r"):

            data = line[:-1].split("\t")

            if len(data) == 2:
                # ignore synonyms of non-existent contigs
                identifier = data[1]
                if data[0] not in self.mIndex: continue
                self.mSynonyms[identifier] = data[0]
            else:
                ## index with random access points
                if len(data) > 4:
                    (identifier, pos_id, block_size, lsequence) = data[0], int(data[1]), int(data[2]), int(data[-1])
                    points = map(int, data[3:-1])
                    self.mIndex[identifier] = (pos_id, block_size, lsequence, points)
                else:
                    (identifier, pos_id, pos_seq, lsequence) = data[0], int(data[1]), int(data[2]), int(data[-1])
                    self.mIndex[identifier] = struct.pack( "QQi", pos_id, pos_seq, lsequence)
                    

        self._addSynonyms()
        self.mIsLoaded = True
        
    def _addSynonyms( self ):
        '''add synonyms to indices.
        '''

        # Treat common cases of naming incompatibilites like chr1 = 1. 
        # Truncate or add known prefixes
        def _add( src, target ):
            for p in ("chr", "contig", "scaffold", "Chr" ):
                if src.startswith(p):
                    k = src[len(p):]
                    if k not in self.mIndex: self.mSynonyms[k]= target
                    # add lower/upper-case version
                    k = src[0].upper() + src[1:]
                    if k not in self.mIndex: self.mSynonyms[k]= target
                    k = src[0].lower() + src[1:]
                    if k not in self.mIndex: self.mSynonyms[k]= target
                    break
            else:
                for p in ("chr", "contig", "scaffold" ):
                    k = "%s%s" % (p,src)
                    if k not in self.mIndex: 
                        self.mSynonyms[k]= target

        k = self.mSynonyms.items()

        # fix the ambiguity between chrMT and chrM between UCSC and ENSEMBL
        if "chrM" in self.mIndex and "chrMT" not in self.mIndex:
            self.mSynonyms[ "chrMT" ] = "chrM"
        elif "chrM" not in self.mSynonyms and "chrMT" in self.mSynonyms:
            self.mSynonyms[ "chrM" ] = "chrMT"

        for key, val in k: _add( key, val )

        # add pointers to self
        for key in self.mIndex.keys(): _add( key, key )

    def setTranslator(self, translator = None ):
        """set the :class:`Translator` to use."""
        self.mTranslator = translator
            
    def getDatabaseName( self ):
        """returns the name of the database."""
        return self.mDbname

    def getToken( self, contig ):
        """check if token is in index."""
        if not self.mIsLoaded: self._loadIndex()

        if contig in self.mSynonyms:
            contig = self.mSynonyms[contig]

        if contig not in self.mIndex:
            raise KeyError, "%s not in index" % contig

        return contig

    def getLength( self, contig ):
        """return sequence length for sbjct_token."""
        if not self.mIsLoaded: self._loadIndex()
        return struct.unpack( "QQi", self.mIndex[self.getToken(contig)] )[2]

    def getLengths( self ):
        """return all sequence lengths."""
        if not self.mIsLoaded: self._loadIndex()
        return [ struct.unpack( "QQi", x)[2] for x in self.mIndex.values() ]

    def compressIndex( self ):
        """compress index.
        Creates a database interface to an index.
        """
        self._loadIndex( compress=True )

    def getContigs( self ):
        """return a list of contigs (no synonyms)."""
        if not self.mIsLoaded: self._loadIndex()
        return self.mIndex.keys()

    def getContigSizes( self, with_synonyms = True ):
        """return hash with contig sizes including synonyms."""
        if not self.mIsLoaded: self._loadIndex()

        contig_sizes = {}
        for key, val in self.mIndex.items():
            contig_sizes[key] = self.getLength( key )

        if with_synonyms:
            for key, val in self.mSynonyms.items():
                contig_sizes[key] = self.getLength( val )

        return contig_sizes

    def setConverter( self, converter ):
        """set converter from coordinate system to 0-based, both strand, open/closed
        coordinate system."""
        self.mConverter = converter

    def getSequence( self,
                     contig, 
                     strand = "+", 
                     start = 0, 
                     end = 0,
                     converter = None,
                     as_array = False):
        """get a genomic fragment.

        A genomic fragment is identified by the coordinates
        contig, strand, start, end.

        The converter function supplied translated these coordinates
        into 0-based coordinates. By default, start and end are assumed
        to be pythonic coordinates and are forward/reverse coordinates.

        If as_array is set to true, return the AString object. This might
        be beneficial for large sequence chunks. If as_array is set to False,
        return a python string.
        """

        contig = self.getToken( contig )

        data = self.mIndex[contig]
        # dummy is
        # -> pos_seq for seekable streams
        # -> block_size for unseekable streams
        try:
            pos_id, dummy, lsequence = struct.unpack( "QQi", data )
        except struct.error:
            pos_id, dummy, lsequence, points = data

        pos_seq = dummy
        block_size = dummy
        
        if end == 0: end = lsequence
        
        if end > lsequence:
            raise ValueError("3' coordinate on %s out of bounds: %i > %i" % (contig, end, lsequence))
        if start < 0:
            raise ValueError("5' coordinate on %s out of bounds: %i < 0" % (contig, start))

        if converter:
            first_pos, last_pos = converter( start, end,
                                             str(strand) in ("+", "1"),
                                             lsequence )
        elif self.mConverter:
            first_pos, last_pos = self.mConverter( start, end,
                                                   str(strand) in ("+", "1"),
                                                   lsequence )
        else:
            first_pos, last_pos = start, end
            if str(strand) in ("-", "0", "-1"):
                first_pos, last_pos = lsequence - last_pos, lsequence - first_pos

        if first_pos == last_pos: return ""

        assert first_pos < last_pos, "first position %i is larger than last position %i " % (first_pos, last_pos)
        
        p = AString()
        
        if self.mNoSeek:
            ## read directly from position
            p.fromstring( self.mDatabaseFile.read( block_size, data[3], first_pos, last_pos) )
        else:
            first_pos += pos_seq
            last_pos += pos_seq

            self.mDatabaseFile.seek( first_pos )
            p.fromstring( self.mDatabaseFile.read( last_pos - first_pos ) )

        if str(strand) in ("-", "0", "-1"):
            p.reverse()            
            p = AString( string.translate( p[:],
                                           string.maketrans("ACGTacgt", "TGCAtgca") ) )

        if self.mTranslator:
            return self.mTranslator.translate( p )
        elif as_array:
            return p
        else:
            # cast to string
            return p[:]

    def getRandomCoordinates( self, size ):
        """returns coordinates for a random fragment of size #.

        The coordinates are forward/reverse.

        Default sampling mode:

        Each residue has the same probability of being
        in a fragment. Thus, the fragment can be smaller than
        size due to contig boundaries.
        """
        if not self.mIsLoaded: self._loadIndex()

        token = random.choice( self.mIndex.keys() )        
        strand = random.choice( ("+", "-") )
        data = self.mIndex[token]
        pos_id, pos_seq, lcontig = struct.unpack( "QQi", data )
        rpos = random.randint( 0, lcontig )
        if size >= lcontig:
            start = 0
            end = lcontig
        else:
            if random.choice( ("True", "False") ):
                start = rpos
                end = min(rpos + size, lcontig)
            else:
                start = max(0, rpos - size)
                end = rpos
            
        return token, strand, start, end


class PysamIndexedFasta( CGATIndexedFasta ):
    '''interface a  pysam/samtools indexed fasta file with the
    CGATIndexedFasta API.'''
    
    def __init__(self, dbname ):

        # open database file and truncate 
        if os.path.exists( dbname ) and dbname.endswith( ".fa" ):
            self.mDatabaseFile = pysam.Fastafile( dbname )
            dbname = dbname[:-len(".fa")]
        elif os.path.exists( dbname ) and dbname.endswith( ".fasta" ):
            self.mDatabaseFile = pysam.Fastafile( dbname )
            dbname = dbname[:-len(".fasta")]
        elif os.path.exists( dbname + ".fa" ):
            self.mDatabaseFile = pysam.Fastafile( dbname + ".fa" )
        elif os.path.exists( dbname + ".fasta" ):
            self.mDatabaseFile = pysam.Fastafile( dbname + ".fasta" )

        if os.path.exists( dbname + ".fai" ):
            self.mNameIndex = dbname + ".fai"
        elif os.path.exists( dbname + ".fa.fai" ):
            self.mNameIndex = dbname + ".fa.fai"
        elif os.path.exists( dbname + ".fasta.fai" ):
            self.mNameIndex = dbname + ".fasta.fai"
        else:
            raise ValueError( "PysamIndexedFasta requires pre-built index")
        
        self.mMethod = "faidx"
        self.mDbname = dbname
        self.mNoSeek = False
        self.mIsLoaded = False
        self.mSynonyms = {} 
        self.mConverter = None
        self.mIndex = {}
        self.mTranslator = None


    def _loadIndex( self, compress = False ):
        '''load index into memory.'''

        with open( self.mNameIndex, "r") as infile:

            for line in infile:
                contig, lsequence, offset, line_blen, line_len = line[:-1].split()
                self.mIndex[contig] = struct.pack( "QQi", int(offset), int(offset), int(lsequence))
                
        self.mIsLoaded = True
        self._addSynonyms()

    def getSequence( self,
                     contig, 
                     strand = "+", 
                     start = 0, 
                     end = 0,
                     converter = None,
                     as_array = False):

        contig = self.getToken( contig )
        
        data = self.mIndex[contig]
        pos_id, dummy, lsequence = struct.unpack( "QQi", data )

        # convert to 0-based positive strand coordinates
        if converter:
            first_pos, last_pos = converter( start, end,
                                             str(strand) in ("+", "1"),
                                             lsequence )
        elif self.mConverter:
            first_pos, last_pos = self.mConverter( start, end,
                                                   str(strand) in ("+", "1"),
                                                   lsequence )
        else:
            first_pos, last_pos = start, end
            if str(strand) in ("-", "0", "-1"):
                first_pos, last_pos = lsequence - last_pos, lsequence - first_pos

        sequence = self.mDatabaseFile.fetch( contig, first_pos, last_pos )
        
        if str(strand) in ("-", "0", "-1"):
            sequence = string.translate( sequence[::-1],
                                         string.maketrans("ACGTacgtNn", "TGCAtgcaNn") )

        return sequence

def IndexedFasta( dbname, *args, **kwargs ):
    '''factory function for IndexedFasta objects.'''

    if (os.path.exists( dbname ) or os.path.exists( dbname + ".fa" )) \
            and ( os.path.exists( dbname + ".fai") or \
                      os.path.exists( dbname + ".fa.fai") ):
        return PysamIndexedFasta( dbname, *args, **kwargs )
    else:
        return CGATIndexedFasta( dbname, *args, **kwargs )

###############################################################################
###############################################################################
###############################################################################
## converter functions. Some code duplication could be avoided but 
## I preferred to keep the functions lean.
###############################################################################        
def _one_forward_closed(x, y, c, l):
    """convert coordinates to zero-based, both strand, open/closed coordinates.
    
     Parameters are from, to, is_positive_strand, length of contig.
    """
    x -= 1
    if not c: x, y = l - y, l - x
    return x, y 
def _zero_forward_closed(x, y, c, l):
    """convert coordinates to zero-based, both strand, open/closed coordinates.
    
     Parameters are from, to, is_positive_strand, length of contig.
    """
    y += 1
    if not c: x, y = l - y, l - x
    return x, y 
def _one_both_closed(x, y, c = None, l = None):
    """convert coordinates to zero-based, both strand, open/closed coordinates.
    
     Parameters are from, to, is_positive_strand, length of contig.
    """
    return x - 1, y
def _zero_both_closed(x, y, c = None, l = None):
    """convert coordinates to zero-based, both strand, open/closed coordinates.
    
     Parameters are from, to, is_positive_strand, length of contig.
    """
    return x, y + 1

def _one_forward_open(x, y, c, l):
    """convert coordinates to zero-based, both strand, open/closed coordinates.
    
     Parameters are from, to, is_positive_strand, length of contig.
    """
    x -= 1
    y -= 1
    if not c: x, y = l - y, l - x
    return x, y 
def _zero_forward_open(x, y, c, l):
    """convert coordinates to zero-based, both strand, open/closed coordinates.
    
     Parameters are from, to, is_positive_strand, length of contig.
    """
    if not c: x, y = l - y, l - x
    return x, y 
def _one_both_open(x, y, c = None, l = None):
    """convert coordinates to zero-based, both strand, open/closed coordinates.
    
     Parameters are from, to, is_positive_strand, length of contig.
    """
    return x - 1, y - 1

def _zero_both_open(x, y, c = None, l = None):
    """convert coordinates to zero-based, both strand, open/closed coordinates.
    
    Parameters are from, to, is_positive_strand, length of contig.
    """
    return x, y

def getConverter( format ):
    """return a converter function for converting various
    coordinate schemes into 0-based, both strand, closed-open ranges.

    converter functions have the parameters
    x, y, s, l: with x and y the coordinates of
    a sequence fragment, s the strand (True is positive)
    and l being the length of the contig.

    Format is a "-" separated combination of the keywords
    "one", "zero", "forward", "both", "open", "closed":

    zero/one: zero or one-based coordinates
    forward/both: forward coordinates or forward/reverse coordinates
    open/closed: half-open intervals (pythonic) or closed intervals

    """

    data = set(format.split("-"))

    if "one" in data:
        if "forward" in data:
            if "closed" in data:
                return _one_forward_closed                
            else:
                return _one_forward_open
        else:
            if "closed" in data:
                return _one_both_closed
            else:
                return _one_both_open
    else:
        if "forward" in data:
            if "closed" in data:
                return _zero_forward_closed
            else:
                return _zero_forward_open
        else:
            if "closed" in data:
                return _zero_both_closed
            else:
                return _zero_both_open                

## Test function for benchmarking purposes
def benchmarkRandomFragment( fasta, size ):
    """returns a random fragment of size."""

    contig, strand, start, end = fasta.getRandomCoordinates( size )
    s = fasta.getSequence( contig, strand, start, end )
    return s

def verify( fasta1, fasta2, num_iterations, fragment_size,
            stdout = sys.stdout, quiet = False ):
    """verify two databases.

    Get segment from fasta1 and check for presence in fasta2.
    """
    if not quiet:
        options.stdout.write("verifying %s and %s using %i random segments of length %i\n" %\
                             (fasta1.getDatabaseName(),
                              fasta2.getDatabaseName(),
                              num_iterations,
                              fragment_size ))
        options.stdout.flush()
    nerrors = 0
    for x in range(num_iterations):
        contig, strand, start, end = fasta1.getRandomCoordinates( fragment_size )
        s1 = fasta1.getSequence(contig,strand,start,end)
        s2 = fasta2.getSequence(contig,strand,start,end)
        if s1 != s2:
            if not quiet:
                options.stdout.write("discordant segment: %s:%s:%i:%i\n%s\n%s\n" %\
                                     (contig, strand, start, end, s1, s2) )
            nerrors += 1
    return nerrors

def splitFasta( infile, chunk_size, dir="/tmp", pattern=None ):
    """split a fasta file into a subset of files.

    If pattern is not given, random file names are chosen.
    """

    n = 0
    chunk = 0
    
    def _getFilename( chunk ):
        if pattern:
            outname = pattern % chunk
            outfile = os.open( outname, "w" )
        else:
            (outfile, outname) = tempfile.mkstemp( dir=dir )
        return (outfile, outname)

    outfile, outname = _getFilename(chunk)
    filenames = [ outname ]
    noutput = 0
    for line in infile:
        if line[0] == "#": continue
        if line[0] == ">":
            n += 1
            if n > chunk_size:
                os.close(outfile)
                n = 1               
                outfile, outname = _getFilename(chunk)
                filenames.append( outname )
            noutput += 1
            
        os.write( outfile, line )
        
    os.close(outfile)

    if noutput == 0:
        os.remove( outname )
        return []

    return filenames

def parseCoordinates( s ):
    '''parse a coordinate string.'''

    if ":" in s:
        d = s.strip().split(":")
        if len(d) == 4:
            contig, strand, start, end = d
        elif len(d) == 2:
            contig = d[0]
            strand = "+"
            if ".." in d[1]:
                start, end = d[1].split("..")
            else:
                start = d[1]
        else:
            raise ValueError("format not recognized")
    else:
        contig = s
        strand = "+"
        start = "0" 
        # full sequence
        end = "0"

    start = int( re.sub( ",", "", start) )
    if end: end = int( re.sub( ",", "", end) )
    else: end = start + 1
    
    return contig, strand, start, end




################################################################################
#   Gene prediction pipeline 
#
#   $Id: tree2stats.py 2782 2009-09-10 11:40:29Z andreas $
#
#   Copyright (C) 2004 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
"""
tree2stats.py - compute statistics on trees
===========================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

Read a collection of trees from stdin and compute statistics on
branchlengths over all trees.

Usage
-----

Example::

   python <script_name>.py --help

Type::

   python <script_name>.py --help

for command line help.

Documentation
-------------

Code
----

""" 
import os
import sys
import string
import re
import getopt
import time
import optparse
import math
import tempfile
import subprocess

from types import *

import CGAT.Experiment as E
import CGAT.TreeTools as TreeTools
import CGAT.Stats as Stats

USAGE="""python tree2stats.py [options] < stdin

compute summary statistics for trees.
"""

if __name__ == "__main__":

    parser = E.OptionParser( version = "%prog version: $Id: tree2stats.py 2782 2009-09-10 11:40:29Z andreas $",
                                    usage = globals()["__doc__"] )

    parser.add_option("-m", "--method", dest="methods", type="choice", action="append",
                      choices=("branchlengths",),
                      help="methods to apply." )
                      
    parser.set_defaults(
        methods = [],
        filtered_branch_length = -999,
        )

    (options, args) = E.Start( parser, add_pipe_options = True )

    nexus = TreeTools.Newick2Nexus( sys.stdin )
    if options.loglevel >= 1:
        options.stdlog.write( "# read %i trees from stdin.\n" % len(nexus.trees))

    ninput = len(nexus.trees)

    nskipped = 0
    
    for method in options.methods:

        outfile = options.stdout
        
        if method == "branchlengths":

            outfile.write("tree\t%s\n" % "\t".join(Stats.DistributionalParameters().getHeaders()))
            
            for tree in nexus.trees:
                branchlengths = []
                for node in tree.chain.values():
                    # ignore branch length of root if it is zero
                    if not node.prev and node.data.branchlength == 0: continue

                    if node.data.branchlength == options.filtered_branch_length: continue
                    
                    branchlengths.append( node.data.branchlength )
                    
                s = Stats.DistributionalParameters( branchlengths )
                outfile.write("%s\t%s\n" % (tree.name, str(s)) )
            
    if options.loglevel >= 1:
        options.stdlog.write( "# ninput=%i, nskipped=%i\n" % (ninput, nskipped))
        
    E.Stop()



#! /bin/env python
################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
csv2db.py - upload table to database
====================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

create a table from a csv separated file and load data into it.

This module supports backends for postgres and sqlite3. Column types are
auto-detected.

Read a table from stdin and create an sqlite3 database. By default, the database
will reside in a file called csvdb and in a table csv.

  
.. todo::

   Use file import where appropriate to speed up loading. Currently, this is
   not always the case.

Usage
-----

Example::

   python csv2db.py -b sqlite < stdin 

Type::

   python csv2db.py --help

for command line help.

Documentation
-------------

Code
----

'''

import os
import sys
import re
import CGAT.Experiment as E
import CGAT.CSV2DB as CSV2DB
import CGAT.CSV as CSV

import csv
import sqlite3

csv.field_size_limit(sys.maxint)

def main( argv = sys.argv ):

    parser = CSV2DB.buildParser()

    (options, args) = E.Start( parser, argv = argv, add_psql_options = True )

    if options.from_zipped:
        import gzip
        infile = gzip.GzipFile( fileobj= options.stdin, mode='r')

    else:
        infile = options.stdin

    CSV2DB.run( infile, options )

    E.Stop()

if __name__ == "__main__":
    sys.exit(main())



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id: snp2table.py 2861 2010-02-23 17:36:32Z andreas $
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
"""
snp2table.py - annotate variants
================================

:Author: Andreas Heger
:Release: $Id: snp2table.py 2861 2010-02-23 17:36:32Z andreas $
:Date: |today|
:Tags: Python

Purpose
-------

This script reads a file in samtools pileup -c format and annotates the
SNPs in that file.

.. note::

   This script is still under construction. There are issues with the
   phasing of variants, using the wild-type sequence and merging
   splice and coding variants.

Usage
-----

Example::

   gunzip < snps.pileup.gz 
   | python snp2table.py  
        --genome-file=genome 
        --filename-exons=exons
        --filename-annotations=bases
        --filename-junctions=junctions
        --log=log 
   > result.out

Type::

   python <script_name>.py --help

for command line help.

Documenation
------------

The following annotations are available:

exons
   Given a gtf-file with the output from ``gtf2gff.py --method=exons``, each
   SNP that overlaps an exon is labeled according to its position within a transcript. 
   The columns are:

   ntranscripts
      number of transrcipts in the gene that the exon belongs to.

   nused
      number of transcripts that contain this exon 
   
   positions
      positions of exons within transcipts

annotation
   Given a fasta file with base-level annotations (output from ``gtf2fasta.py``),
   each SNP is labeled by its effect. The columns are:

   code
      one letter annotation of base. See :file:`gtf2fasta.py`.
   reference_codon
      the reference codon (in case sequence is coding)
   reference_aa
      the reference amina acid (in case sequence is coding)
   variant_type
      the variant types ('I' for insertion, 'D' for deletion', 'S' for substitution).
   variant_codon
      the variant codons (in case sequence is coding)
   variant_aa
      the variant amino acids (in case sequence is coding)

   Variants will be separated by ``,`` if there are several variants per position.

The script deconvolutes substitution and indel snps. The following columns are always output:

chromosome
   chromosome
position
   reference base position  (0-based)
reference_base
   reference base. For indels, this is ``*``.
genotype
   consensus base. For indels, this is the genotype
consensus_quality
   consensus quality
snp_quality
   snp quality
rms_mapping_qualit
   rms quality
coverage
   coverage
read_bases
   read bases. For indels, this is the first allele (``*`` for reference).
base_qualities
   base qualities. For indels, this is the second allele (``*`` for reference).

See the samtools documentation for the meaning of these.

Code
----

""" 

import os
import sys
import re
import optparse
import collections

import CGAT.Experiment as E

import pysam
import CGAT.IndexedFasta as IndexedFasta
import CGAT.IndexedGenome as IndexedGenome
import CGAT.Genomics as Genomics

def readJunctions( filename_junctions ):
    '''read junctions from a tab-separated file.

    returns a dictionary of dictionaries for splice junctions per
    contig and position.
    '''
    
    junctions = collections.defaultdict( dict )
    infile = open( filename_junctions, "r" )
    njunctions = 0

    for line in infile:
        if line.startswith( "#" ): continue
        data = line[:-1].split("\t")
        if data[0] == "contig": continue
        contig, strand, end, start, frame = data[:5]
        start,end,frame = int(start), int(end), int(frame)
        j = junctions[contig]
        if end in j:
            j[end].add( (strand,start,frame) )
        else:
            j[end] = set( ((strand,start,frame),) )
        njunctions += 1

    infile.close()
    E.info( "read %i junctions for %i contigs" % (njunctions, len(junctions) ) )
    return junctions

class BaseAnnotator(object):
    '''annotator for single bases in the genome.'''

    mHeader = ()

    def __init__(self, fasta = None, *args, **kwargs):
        self.mFasta = fasta

    def __str__( self ):
        return ""

    def getHeader( self ):
        '''return header'''
        return "\t".join(self.mHeader)

class BaseAnnotatorSNP( BaseAnnotator ):
    '''output SNPs. 
    
    This annotator deconvolutes substitution and indel SNPs.
    '''

    mHeader = (
        "chromosome", 
        "position", 
        "reference_base", 
        "genotype",
        "consensus_quality",
        "snp_quality",
        "rms_mapping_quality",
        "coverage" )

    mAdditionalHeader = (
        "read_bases",
        "base_qualities" )

    def __init__(self, *args, **kwargs):
        BaseAnnotator.__init__(self, *args,**kwargs)
    
    def __str__(self ):
        # truncate the last two columns to make the snp output even length
        return "\t".join( map(str, self.mSNP)[:len(self.mHeader)] )

    def update( self, snp ):
        '''update with snp.'''
        self.mSNP = snp

class BaseAnnotatorExons( BaseAnnotator ):
    '''annotate SNP by exons that it overlaps with.'''

    mHeader = [ "exons_%s" % x for x in ( "ntranscripts", "nused", "pos" )]

    def __init__(self, filename_exons, *args, **kwargs):
        BaseAnnotator.__init__(self, *args,**kwargs)

        exons = IndexedGenome.IndexedGenome()
        nexons = 0
        for g in GTF.iterator( open( filename_exons, "r") ):
            exons.add( g.contig, g.start, g.end, g )
            nexons += 1

        self.mExons = exons

        E.info( "indexed %i exons on %i contigs" % (nexons, len(exons) ) )

    def update( self, snp ):
        '''update with snp.'''

        exons = list(self.mExons.get( snp.chromosome, 
                                      snp.pos, 
                                      snp.pos+1))

        if exons:
            for start, end, gff in exons:
                self.mOTranscripts = "%i" % gff["ntranscripts"]
                self.mOUsed = "%i" % gff["nused"]
                self.mOPos = "%s" % gff["pos"]
        else:
            self.mOTranscripts = "na"
            self.mOUsed = "na"
            self.mOPos = "na"

    def __str__(self):
        return "\t".join( (self.mOTranscripts, self.mOUsed, self.mOPos) )

class BaseAnnotatorSpliceSites( BaseAnnotator ):
    '''annotate SNP by splice sites that it overlaps with.

    Splice sites are read from filename_junctions.

    The module is aware of the following splice signals
    
       * GT/AG: A
       * GC/AG: B
       * AT/AC: C
       
    A SNP is labeled by splice site if it is directly overlapping.
    Furthermore, it will be labeled as synonymous (S), non-synomous (N,
    if it changes from one splice class to another) or as a deleterious (D).

    In particular, the codes are:
    
+---------------+----------------------------------------------------------------------+                                                                                                                                                                   
|code           | description                                                          |
+---------------+----------------------------------------------------------------------+                                                                                                                                                                   
|S              | synonymous SNP - the splice signal is unchanged                      |
+---------------+----------------------------------------------------------------------+
|N              | non-synonymous SNP - one splice site changes to another              |
+---------------+----------------------------------------------------------------------+
|D              | deleterious snp, a canonical  splice site is abrogated               |
+---------------+----------------------------------------------------------------------+
|U              | change from unknown to unknown splice                                |
+---------------+----------------------------------------------------------------------+
|C              | creation of an new canonical splice site                             |
+---------------+----------------------------------------------------------------------+


    All SNPS will be annotatoted if they are within 10 bases of a splice
    site. Furthermore, the overlap flag will be set for those
    SNPs overlapping one of the known splice signals.

    Note that a disrupted splice signal might still lead to a valid
    protein product.
    '''

    mHeader = [ "splice_%s" % x \
                for x in ( "transcript_ids", "gene_ids", "has_overlap", "is_homozygous", "old", "new", "distance", "code" )]
    mIntronTypes = ( ("U2-GT/AG", "GT", "AG" ),
                     ("U2-nc-GC/AG", "GC", "AG" ),
                     ("U12-AT/AC", "AT", "AC") )
    # size of splice signal
    mSize = 10

    def __init__(self, filename_junctions, *args, **kwargs):
        BaseAnnotator.__init__(self, *args,**kwargs)

        junctions = IndexedGenome.IndexedGenome()
        
        infile = open( filename_junctions, "r" )
        njunctions = 0

        for line in infile:
            if line.startswith( "#" ): continue
            data = line[:-1].split("\t")
            if data[0] == "contig": continue
            # end, start are the positions of the last base of the codon
            # 5' of the intron and first base of codon 3' of the intron.
            contig, strand, end, start, frame, gene_id, transcript_id = data
            start,end,frame = int(start), int(end), int(frame)
            # convert to intron coordinates
            intron_start, intron_end = end + 1, start
            # convert to positive strand coordinates
            if strand == "-":
                lcontig = self.mFasta.getLength( contig )
                intron_start,intron_end = lcontig-intron_end, lcontig-intron_start
            junctions.add( contig, intron_start, intron_start+self.mSize, (strand, intron_start, intron_end, gene_id, transcript_id) )
            junctions.add( contig, intron_end-self.mSize, intron_end, (strand, intron_start, intron_end, gene_id, transcript_id) )
            njunctions += 1
        infile.close()
        
        self.mJunctions = junctions
        
        E.info( "read and indexed %i junctions for %i contigs" % (njunctions, len(junctions) ) )

    def hasSpliceMotif( self, seq5, seq3 ):
        """find a splice motif in seq.

        returns the name of the splice motif or None
        """
        for name, prime5, prime3 in self.mIntronTypes:
            if seq5.startswith( prime5 ) and seq3.endswith( prime3 ):
                return name, prime5, prime3
        return None, None, None

    def buildSequenceVariants( self, seq, strand, pos, snp ):
        '''build new sequence by modifying a sequence fragment in seq at
        pos with snp.

        It is assumed that seq is already oriented according to strand.
        The strand is used to revert the snp if necessary.

        Note that only sequences different from seq will be returned.

        returns is_homozygous, seqs
        '''
        is_negative_strand = Genomics.IsNegativeStrand( strand )
        reference_base = snp.reference_base

        if reference_base != "*" and is_negative_strand:
            reference_base = Genomics.complement( reference_base )

        new_sequences = []
        is_homozygous = True
        if reference_base != "*":
            if seq[pos].upper() != reference_base.upper():
                raise ValueError( "base mismatch at snp %i, expected %s, got %s in %s at position %i; snp=%s" % \
                                  (snp.pos, reference_base, seq[pos], seq, pos,
                                   ";".join(map(str, snp))) )

            # single base changes
            variant_bases = Genomics.resolveAmbiguousNA( snp.genotype )
            if len(variant_bases) == 1:
                is_homozygous = True
            else:
                is_homozygous = False

            for variant_base in variant_bases:
                if is_negative_strand:
                    variant_base = Genomics.complement( variant_base )

                s = list(seq)
                s[pos] = variant_base
                s = "".join(s)
                if s != seq: new_sequences.append( s )
        else:
            variants = snp.genotype.split("/")
            is_homozygous = False
            for variant in variants:
                
                s = list(seq)
                # samtools denotes insert/deletion after position
                # while python is before/at position, hence the pos+1
                if variant[0] == "+":
                    toinsert = variant[1:].upper()
                    if is_negative_strand:
                        toinsert = Genomics.complement( toinsert )
                        s.insert( pos, toinsert )
                    else:
                        s.insert( pos+1, toinsert )

                elif variant[0] == "-":
                    # pos+1+len(x)-1 = pos+len(x)
                    todelete = variant[1:].upper()
                    l = len(todelete)
                    if is_negative_strand:
                        # delete left of pos
                        xstart = max(0, pos-l)
                        xend = pos
                        todelete = todelete[:min(l,pos)]
                    else:
                        # delete right of pos
                        xstart = pos+1
                        xend = min(self.mSize, pos + 1 + l)
                        todelete = todelete[:self.mSize-(pos+1)]
                            
                    deleted = "".join(s[xstart:xend])
                    
                    if is_negative_strand:
                        deleted = Genomics.complement( deleted )

                    if deleted != todelete:
                        raise ValueError( "base mismatch at indel %i, expected %s, got %s in %s at position %i(%i:%i); is_negative_strand=%s, snp=%s" % \
                                (snp.pos, todelete, deleted, seq, pos, xstart, xend,
                                 is_negative_strand,
                                 ";".join(map(str, snp))) )
                    del s[xstart:xend]
                    
                elif variant[0] == "*":
                    is_homozygous = True
                else:
                    raise ValueError( "unknown variant sign '%s'" % variant[0] )

                s = "".join(s)
                if s != seq: new_sequences.append( s )
            
        return is_homozygous, new_sequences

    def update( self, snp ):
        '''update with snp.'''

        junctions = list(self.mJunctions.get( snp.chromosome, 
                                              snp.pos, 
                                              snp.pos+1))

        self.mResults = []

        reference_base = snp.reference_base
        
        if junctions:
            for start, end, data in junctions:
                strand, intron_start, intron_end, gene_id, transcript_id = data
                snp_pos = snp.pos
                
                if strand == "-":
                    lcontig = self.mFasta.getLength( snp.chromosome )
                    intron_start, intron_end = lcontig - intron_end, lcontig - intron_start
                    snp_pos = lcontig - snp_pos - 1

                seq5 = self.mFasta.getSequence( snp.chromosome, strand, intron_start, intron_start + self.mSize ).upper()
                seq3 = self.mFasta.getSequence( snp.chromosome, strand, intron_end - self.mSize, intron_end ).upper()

                splice_name, prime5, prime3 = self.hasSpliceMotif( seq5, seq3 )
                if splice_name:
                    if intron_start <= snp_pos < intron_start + len(prime5):
                        has_overlap = "5"
                    elif intron_end - len(prime3) <= snp_pos < intron_end:
                        has_overlap = "3"
                    else:
                        has_overlap = ""
                else:
                    has_overlap = ""
                    
                E.debug( "building a splice site for %s:%s:%i, splice=%i:%i:%s (snp=%i), splice_name=%s, has_overlap=%s" %\
                         (snp.chromosome, strand, snp.pos,
                          intron_start, intron_end, strand, snp_pos,
                          str(splice_name),
                          str(has_overlap) ))

                if intron_start <= snp_pos < intron_start + self.mSize:
                    is_homozygous, new_seq5 = self.buildSequenceVariants( seq5, strand, snp_pos - intron_start, snp )
                    new_seqs = [ (x, seq3) for x in new_seq5 ]
                    distance = snp_pos - intron_start
                elif intron_end - self.mSize <= snp_pos < intron_end:
                    is_homozygous, new_seq3 = self.buildSequenceVariants( seq3, strand, self.mSize - (intron_end - snp_pos), snp )
                    new_seqs = [ (seq5, x) for x in new_seq3 ]
                    distance = intron_end - snp_pos - 1
                else:
                    raise ValueError( "no overlap?")

                E.debug( "building a splice site for %s:%s:%i, splice=%i:%i:%s, splice_name=%s, distance=%i, has_overlap=%s, is_homozygous=%s, new_seqs=%s" %\
                         (snp.chromosome, strand, snp.pos,
                          intron_start, intron_end, strand,
                          str(splice_name),
                          distance,
                          str(has_overlap),
                          str(is_homozygous),
                          str(new_seqs) ) )

                # indels at the end of size are out-of-range - ignore
                if new_seqs == []: continue

                for new_seq5, new_seq3 in new_seqs:
                    new_splice_name, new_prime5, new_prime3 = self.hasSpliceMotif( new_seq5, new_seq3 )

                    if splice_name == None and new_splice_name == None:
                        code = "U"
                    elif new_splice_name == None:
                        code = "D"
                    elif splice_name == None:
                        code = "C"
                    elif splice_name == new_splice_name:
                        code = "S"
                    elif splice_name != new_splice_name:
                        code = "N"

                    if is_homozygous: is_homozygous = "1"
                    else: is_homozygous = "0"
                    if not splice_name: splice_name = ""
                    if not new_splice_name: new_splice_name = ""
                    self.mResults.append( (transcript_id, gene_id, has_overlap, is_homozygous,
                                           splice_name, new_splice_name, str(distance), code ) )

    def __str__(self):
        if self.mResults:
            return "\t".join( [ ";".join( x ) for x in zip( *self.mResults ) ] )
        else:
            return "\t" * 8
        
class BaseAnnotatorCodon( BaseAnnotator ):
    '''annotate snp by coding sequence that it overlaps with

    Variant types are annotated with a one-letter code.

+---------------+----------------------------------------------------------------------+                                                                                                                                                                   
|code           | description                                                          |
+---------------+----------------------------------------------------------------------+                                                                                                                                                                   
|O              | homozygous substitution                                              |                                                                                                                                                                   
+---------------+----------------------------------------------------------------------+                                                                                                                                                                   
|E              | heterozygous substitution                                            |
+---------------+----------------------------------------------------------------------+
|I              | insertion with respect to reference genome                           |
+---------------+----------------------------------------------------------------------+
|D              | deletion with respect to reference genome                            |
+---------------+----------------------------------------------------------------------+
|W              | wild type (reference genome)                                         |
+---------------+----------------------------------------------------------------------+

    This function will not annotate indels accurately.
    '''

    mHeader = [ "code", "reference_codon", "reference_aa", "variant_type", "variant_codon", "variant_aa" ]

    def __init__(self, annotations_file, junctions, *args, **kwargs):

        BaseAnnotator.__init__(self, *args,**kwargs)

        self.mAnnotations = IndexedFasta.IndexedFasta( annotations_file )
        self.mJunctions = junctions

    def updateSNPs( self, snp, is_negative_strand, pos ):
        '''update SNPs.'''

        contig = snp.chromosome
        lcontig = self.mFasta.getLength( contig )
        reference_base = snp.reference_base

        if snp.genotype in 'ACGTacgt':
            # homozygous substitution
            self.mVariantType.append( "O" )
        else:
            # heterozygous substitution
            self.mVariantType.append( "E" )

        # switch reference strand codon to correct strand
        if reference_base != "*" and is_negative_strand:
            reference_base = Genomics.complement( reference_base )

        # collect all possible variants of reference codons
        for reference_codon in self.mReferenceCodons:

            self.mReferenceAAs.append( Genomics.translate( reference_codon ) )

            # process single base changes
            variant_bases = Genomics.resolveAmbiguousNA( snp.genotype )

            if reference_codon[ pos ] != reference_base:
                raise ValueError( "base mismatch at %i (codon=%s,%i): codon:%s != genome:%s; `%s`" % \
                                      (snp.pos, reference_codon, pos, reference_codon[pos], reference_base, ";".join(map(str, snp))) )

            for variant_base in variant_bases:
                if is_negative_strand:
                    variant_base = Genomics.complement( variant_base )

        self.mVariantAAs.extend( [ Genomics.translate( x ) for x in self.mVariantCodons ] )

    def updateIndels( self, snp, is_negative_strand ):

        contig = snp.chromosome
        lcontig = self.mFasta.getLength( contig )

        # get location of insertion/deletion. The location
        # is after position, hence get position and position + 1
        code = self.mAnnotations.getSequence( contig, "+", snp.pos, snp.pos+2 )
        self.mCode = code

        variants = snp.genotype.split("/")
        for variant in variants:

            if variant[0] == "*":
                self.mVariantType.append( "W" )

            elif variant[0] == "+":
                toinsert = variant[1:] 
                self.mVariantType.append( "I" )
                
            elif variant[0] == "-":
                todelete = variant[1:]
                # deletions need to be looked at in a wider range
                self.mVariantType.append( "D" )
                
            else:
                raise ValueError( "unknown variant sign '%s'" % variant[0] )


        # ignore non-coding Indels
        if code[0] and code[1] not in 'abcABC': return

        if is_negative_strand:
            variants = [ Genomics.complement(x) for x in variants ]

        for reference_codon in self.mReferenceCodons:

            variants = snp.genotype.split("/")
            variants = [ x[1:] for x in variants ]

            for variant in variants:
                if len(variant) % 3 != 0:
                    self.mVariantCodons.append( "!" )
                else:
                    self.mVariantCodons.append( variant )

            self.mVariantAAs.extend( [ Genomics.translate( x ) for x in self.mVariantCodons ] )

    def update( self, snp ):
        '''update with snp.'''

        contig = snp.chromosome
        lcontig = self.mFasta.getLength( contig )

        # check for codon:
        self.mReferenceCodons = []
        self.mVariantCodons = []
        self.mReferenceAAs = []
        self.mVariantAAs = []
        self.mVariantType = []

        if snp.pos >= lcontig:
            E.warn("snp is out of range: %s:%i => %s" % (contig, lcontig, ";".join( map(str, snp)) ))
            self.mCode = "X"
            return

        reference_base = snp.reference_base

        ######
        code = self.mAnnotations.getSequence( contig, "+", snp.pos, snp.pos+1 )
        self.mCode = code

        is_negative_strand = False

        pos = None
        # ignore non-coding SNPs
        if code in 'abcABC': 

            pos = "abcABC".find(code)
            is_negative_strand = code in 'ABC'

            if is_negative_strand:
                pos -= 3
                position = lcontig - snp.pos - 1
                strand = "-"
            else:
                position = snp.pos
                strand = "+"

            # start, end are forward/negative strand coordinates
            offset = pos % 3 
            start, end = position - offset, position - offset + 3

            # If a codon is split by an intron, several 
            # reference codons are possible. 
            self.mReferenceCodons = set()
            if self.mJunctions:
                if contig in self.mJunctions:
                    j = self.mJunctions[contig]
                    for x in xrange(start, end):
                        if x in j:
                            # found a junction within codon
                            for junction_strand, next_start, frame in j[x]:

                                # wrong strand
                                if (junction_strand == "-") != is_negative_strand:
                                    continue

                                # codon is complete, junction within frame
                                if frame == 0:
                                    self.mReferenceCodons.add( self.mFasta.getSequence( contig, strand, start, end ).upper() )                                    
                                    continue

                                codon = self.mFasta.getSequence( contig, strand, start, x + 1) + \
                                    self.mFasta.getSequence( contig, strand, next_start, next_start + end - (x+1))

                                # position within codon
                                y = x + 1 - start
                                E.debug( "building a split codon for %s:%s:%i(%i) start=%i, split=%i, cont=%i, end=%i: codon=%s:%s" % \
                                         (contig, strand, snp.pos, position, start, x+1, next_start, next_start + end - (x+1),
                                         codon[:y], codon[y:]))

                                assert len(codon) == 3
                                self.mReferenceCodons.add( codon.upper() )

            # ordinary codon                
            if not self.mReferenceCodons:
                self.mReferenceCodons.add( self.mFasta.getSequence( contig, strand, start, end ).upper() )

            # fixate the order of reference codons
            self.mReferenceCodons = tuple( self.mReferenceCodons )

        # process according to variant type
        # indels need to be treated differently from SNPs as
        # they have larger effects
        if reference_base == "*":
            self.updateIndels( snp, is_negative_strand )
        else:
            self.updateSNPs( snp, is_negative_strand, pos )

    def __str__(self):
        return "\t".join( (self.mCode, 
                           ",".join(self.mReferenceCodons), 
                           ",".join(self.mReferenceAAs),
                           ",".join(self.mVariantType), 
                           ",".join(self.mVariantCodons), 
                           ",".join(self.mVariantAAs), ) )

def main( argv = None ):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv: argv = sys.argv

    # setup command line parser
    parser = E.OptionParser( version = "%prog version: $Id: snp2table.py 2861 2010-02-23 17:36:32Z andreas $", usage = globals()["__doc__"] )

    parser.add_option("-g", "--genome-file", dest="genome_file", type="string",
                      help="filename with genome [default=%default]."  )
    parser.add_option("-a", "--filename-annotations", dest="filename_annotations", type="string",
                      help="filename with base annotations (output from gtf2fasta.py) [default=%default]."  )
    parser.add_option("-f", "--filename-exons", dest="filename_exons", type="string",
                      help="filename with exon information (gff formatted file)  [default=%default]."  )
    parser.add_option("-j", "--filename-junctions", dest="filename_junctions", type="string",
                      help="filename with junction information (filename with exon junctions)  [default=%default]."  )
    parser.add_option("-c", "--filename-vcf", dest="filename_vcf", type="string",
                      help="vcf file to parse [default=%default]."  )
    parser.add_option("-i", "--input-format", dest="input_format", type="choice",
                      choices = ("pileup", "vcf" ),
                      help="input format [default=%default]."  )
    parser.add_option( "--vcf-sample", dest="vcf_sample", type="string",
                      help="sample id in vcf file to analyse [default=%default]."  )

    parser.set_defaults(
        genome_file = None,
        filename_annotations = None,
        filename_exons = None,
        filename_junctions = None,
        input_format = "pileup",
        vcf_sample = None,
        filename_vcf = None,
        )

    ## add common options (-h/--help, ...) and parse command line 
    (options, args) = E.Start( parser, argv = argv )

    ninput, nskipped, noutput = 0, 0, 0

    if options.genome_file:
        fasta = IndexedFasta.IndexedFasta( options.genome_file )
    else:
        fasta = None

    if options.filename_junctions:
        junctions = readJunctions( options.filename_junctions )
    else:
        junctions = None

    # setup iterator
    if options.input_format == "pileup":
        iterator = pysam.Pileup.iterate(sys.stdin)
    elif options.input_format == "vcf":
        if not options.vcf_sample:
            raise ValueError( "vcf format requires sample id (--vcf-sample) to be set" )
        if not options.filename_vcf:
            raise ValueError( "reading from vcf requires vcf filename (--filename-vcf) to be set)" )

        iterator = pysam.Pileup.iterate_from_vcf(options.filename_vcf, options.vcf_sample )
    

    modules = []
    modules.append( BaseAnnotatorSNP() )

    if options.filename_exons:
        modules.append( BaseAnnotatorExons( options.filename_exons, fasta=fasta ) )
    if options.filename_annotations:
        modules.append( BaseAnnotatorCodon( options.filename_annotations, fasta=fasta, junctions=junctions ) )
    if options.filename_junctions:
        modules.append( BaseAnnotatorSpliceSites( options.filename_junctions, fasta=fasta ) )
                        
    options.stdout.write( "\t".join( [x.getHeader() for x in modules]) + "\n" )

    for snp in iterator:
        ninput += 1

        # translate chromosome according to fasta
        if fasta: 
            try:
                snp = snp._replace( chromosome=fasta.getToken( snp.chromosome ))
            except KeyError:
                E.warn( "unknown contig `%s` for snp `%s`" % (snp.chromosome, str(snp) ) )
                continue

        for module in modules:
            module.update( snp )

        options.stdout.write( "\t".join( map(str, modules) ) + "\n" )

        noutput += 1

    E.info( "ninput=%i, noutput=%i, nskipped=%i" % (ninput, noutput,nskipped) )

    ## write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit( main( sys.argv) )




################################################################################
#
#   Gene prediction pipeline 
#
#   $Id: gtf2gtf.py 2861 2010-02-23 17:36:32Z andreas $
#
#   Copyright (C) 2004 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
gtf2gtf.py - manipulate transcript models
=========================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This script reads a gene set in :term:`gtf` format from stdin,
applies some transformation, and outputs a new geneset in
:term:`gtf` format to stdout.

Typical operations are:

* merging exons/introns within a transcript/gene, such as:
   ``--merge-exons``, ``--merge-transcripts``, ``--merge-genes``,
   ``--exons2introns``. ``--join-exons``, ``--intersect-transcripts``,
   ``--merge-introns``.

* sorting the input (``--sort``)

* filtering/sampling transcripts/genes (``--filter``, ``--sample-size``,
   ``--remove-overlapping``)

* renaming transcripts/genes (``--unset-genes``, ``--set-protein-to-transcript``,
  ``--set-gene-to-transcript``, --add-protein-id``, ``--set-transcript-to-gene``.
  ``--rename``, ``--renumber-genes``, ``--renumber-transcripts``, 
  ``--transcript2genes``)

* others, such ass ``--reset-strand``.

Most of the transformations require the input to be sorted 
by gene and postion.

Usage
-----

Example::

   python gtf2gtf.py --merge-exons < in.gtf > out.gtf

Type::

   python gtf2gtf.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import optparse
import types
import random
import collections
import itertools

import CGAT.GTF as GTF
import CGAT.Experiment as E
import CGAT.IndexedFasta as IndexedFasta
import CGAT.Genomics as Genomics
import CGAT.Intervals as Intervals
import CGAT.IOTools as IOTools
import CGAT.Components as Components

##------------------------------------------------------------
## This script needs some attention.
##------------------------------------------------------------
def main( argv = None ):

    if not argv: argv = sys.argv

    parser = E.OptionParser( version = "%prog version: $Id: gtf2gtf.py 2861 2010-02-23 17:36:32Z andreas $", 
                             usage = globals()["__doc__"])

    parser.add_option("-m", "--merge-exons", dest="merge_exons", action="store_true",
                      help="merge overlapping exons of all transcripts within a gene. "
                           "The merged exons will be output. "
                           "Input needs to sorted by gene [default=%default]."  )

    parser.add_option("-t", "--merge-transcripts", dest="merge_transcripts", action="store_true",
                      help="merge all transcripts within a gene. The entry will span the whole gene (exons and introns). "
                           "The transcript does not include the UTR unless --with-utr is set. [default=%default]."  )

    parser.add_option( "--merge-genes", dest="merge_genes", action="store_true",
                       help="merge overlapping genes if their exons overlap. "
                       "A gene with a single transcript containing all exons "
                       "of the overlapping transcripts will be output. "
                       "This operation ignores strand information "
                       "The input needs te sorted by transcript [default=%default]."  )

    parser.add_option( "--merge-exons-distance", dest="merge_exons_distance", type="int",
                      help="distance in nucleotides between exons to be merged [default=%default]."  )

    parser.add_option("-j", "--join-exons", dest="join_exons", action="store_true",
                      help="join all exons per transcript. A new transcript will be "
                           "output that spans a whole transcript. "
                           "Input needs to be sorted by transcript [default=%default]."  )

    parser.add_option( "--unset-genes", dest="unset_genes", type="string",
                       help="unset gene identifiers, keeping transcripts intact. "
                       "New gene identifiers are set to the pattern given. For example, "
                       "'--unset-genes=%06i' [default=%default]."  )

    parser.add_option( "--sort", dest="sort", type="choice",
                       choices=("gene", 
                                "transcript", 
                                "position", 
                                "contig+gene", 
                                "position+gene", 
                                "gene+position" ),
                       help="sort input data [default=%default]."  )

    parser.add_option("-u", "--with-utr", dest="with_utr", action="store_true",
                      help="include utr in merged transcripts [default=%default]."  )


    parser.add_option( "--intersect-transcripts", dest="intersect_transcripts", action="store_true",
                      help="intersect all transcripts within a gene. The entry will only span those bases "
                      "that are covered by all transcrips."
                      "The transcript does not include the UTR unless --with-utr is set. This method"
                      "will remove all other features (stop_codon, etc.) "
                      "The input needs to be sorted by gene. "
                      "[default=%default]."  )

    parser.add_option("-i", "--merge-introns", dest="merge_introns", action="store_true",
                      help="merge and output all introns within a gene. The output will contain "
                      "all intronic regions within a gene. Single exon genes "
                      "are skipped. "
                      "The input needs to be sorted by gene. "
                      "[default=%default]."  )

    parser.add_option("-g", "--set-transcript-to-gene", "--set-transcript2gene", dest="set_transcript2gene", action="store_true",
                      help="set the transcript_id to the gene_id [default=%default]."  )

    parser.add_option( "--set-protein-to-transcript", dest="set_protein2transcript", action="store_true",
                      help="set the protein_id to the transcript_id [default=%default]."  )

    parser.add_option( "--add-protein-id", dest="add_protein_id", type="string",
                      help="add a protein_id for each transcript_id. "
                           "The argument is a filename containing a mapping between "
                           "transcript_id to protein_id [default=%default]."  )

    parser.add_option("-G", "--set-gene-to-transcript", "--set-gene2transcript", dest="set_gene2transcript", action="store_true",
                      help="set the gene_id to the transcript_id [default=%default]."  )

    parser.add_option("-d", "--set-score2distance", dest="set_score2distance", action="store_true",
                      help="set the score field for each feature to the distance to "
                           "transcription start site [default=%default]."  )

    parser.add_option( "--exons2introns", dest="exons2introns", action="store_true",
                       help="for each gene build an 'intronic' transcript "
                       "containing the union of all intronic regions "
                       "of all transcripts in a gene."
                       "The features are labeled as 'intron'." 
                       "The input needs to be sorted by gene. "
                      "[default=%default]."  )

    parser.add_option("-f", "--filter", dest="filter", type="choice",
                      choices=("gene", "transcript", "longest-gene", "longest-transcript", "representative-transcript" ),
                      help="apply a filter to the input file. Available filters are: "
                      "'gene-id': filter by gene_id, "
                      "'transcript-id': filter by transcript_id, "
                      "'longest-gene': output the longest gene for overlapping genes ," 
                      "'longest-transcript': output the longest transcript per gene," 
                      "'representative-transcript': output the representative transcript per gene. "
                      "The representative transcript is the transcript that shares most exons with "
                      "the other transcripts in a gene. "
                      "The input needs to be sorted by gene. "
                      "[default=%default]."  )
                      

    parser.add_option("-r", "--rename", dest="rename", type="choice",
                      choices=("gene", "transcript","longest-gene"),
                      help="rename genes or transcripts with a map given by the option `--apply`. "
                      "Those that can not be renamed are removed "
                      "[default=%default]."  )

    parser.add_option( "--renumber-genes", dest="renumber_genes", type="string", 
                       help="renumber genes according to the given pattern. "
                       "[default=%default]."  )
    
    parser.add_option( "--renumber-transcripts", dest="renumber_transcripts", type="string", 
                       help="renumber transcripts according to the given pattern. "
                       "[default=%default]."  )

    parser.add_option("-a", "--apply", dest="filename_filter", type="string",
                      metavar="tsv",
                      help="filename of ids to map/filter [default=%default]." )
    
    parser.add_option( "--invert-filter", dest="invert_filter", action="store_true",
                       help="when using --filter, invert selection (like grep -v). "
                       "[default=%default]."  )

    parser.add_option("--sample-size", dest="sample_size", type="int",
                      help="extract a random sample of size # if the option "
                      "'--filter' is set[default=%default]." )

    parser.add_option("--intron-min-length", dest="intron_min_length", type="int",
                      help="minimum length for introns (for --exons2introns) "
                      "[default=%default]."  )

    parser.add_option("--min-exons-length", dest="min_exons_length", type="int",
                      help="minimum length for gene (sum of exons) (--sample-size) [default=%default]."  )

    parser.add_option("--intron-border", dest="intron_border", type="int",
                      help="number of residues to exclude at intron at either end "
                      "(--exons2introns) [default=%default]."  )

    parser.add_option( "--transcripts2genes", dest="transcripts2genes", action="store_true",
                       help="cluster overlapping transcripts into genes." )

    parser.add_option( "--reset-strand", dest="reset_strand", action="store_true",
                       help="remove strandedness of features (set to '.') when using --transcripts2genes"
                       "[default=%default]."  )

    parser.add_option( "--remove-overlapping", dest="remove_overlapping", type="string",
                       metavar = "gff",
                       help="remove all transcripts that overlap intervals "
                       "in a gff-formatted file."
                       "The comparison ignores strand "
                       "[default=%default]." )

    parser.add_option( "--permit-duplicates", dest="strict", action="store_false",
                       help="permit duplicate genes. "
                       "[default=%default]" )

    parser.add_option( "--remove-duplicates", dest="remove_duplicates", type="choice",
                       choices=("gene", "transcript", "ucsc", "coordinates"),
                       help="remove duplicates by gene/transcript. "
                       "If ``ucsc`` is chosen, transcripts ending on _dup# are "
                       "removed. This is necessary to remove duplicate entries "
                       "that are next to each other in the sort order [%default]" )

    parser.set_defaults(
        sort = None,
        merge_exons = False,
        join_exons = False,
        merge_exons_distance = 0,
        merge_transcripts = False,
        set_score2distance = False,
        set_gene2transcript = False,
        set_transcript2gene = False,
        set_protein2transcript = False,
        add_protein_id = None,
        filename_filter = None,
        filter = None,
        exons2introns = None,
        merge_genes = False,
        intron_border = None,
        intron_min_length = None,
        sample_size = 0,
        min_exons_length = 0,
        transripts2genes = False,
        reset_strand = False,
        with_utr = False,
        invert_filter = False,
        remove_duplications = None,
        remove_overlapping = None,
        renumber_genes = None,
        unset_genes = None,
        renumber_transcripts = None,
        strict = True,
        intersect_transcripts = False,
        )

    (options, args) = E.Start( parser, argv = argv )
    
    ninput, noutput, nfeatures, ndiscarded = 0, 0, 0, 0

    if options.set_transcript2gene:

        for gff in GTF.iterator(options.stdin):

            ninput += 1

            gff.setAttribute( "transcript_id", gff.gene_id)
            options.stdout.write( "%s\n" % str(gff) )                

            noutput += 1
            nfeatures += 1

    elif options.remove_duplicates:

        counts = collections.defaultdict(int)

        if options.remove_duplicates == "ucsc":
            store = []
            remove = set()
            f = lambda x: x[0].transcript_id

            gffs = GTF.transcript_iterator(GTF.iterator(options.stdin), strict = False)
            outf = lambda x: "\n".join( [ str(y) for y in x] )

            for entry in gffs:
                ninput += 1
                store.append(entry)
                id = f(entry)
                if "_dup" in id: 
                    remove.add(re.sub("_dup\d+","", id) )
                    remove.add( id )

            for entry in store:
                id = f(entry)
                if id not in remove: 
                    options.stdout.write( outf(entry) + "\n" )
                    noutput += 1
                else:
                    ndiscarded += 1
                    E.info("discarded duplicates for %s" % (id))
        else:
        
            if options.remove_duplicates == "gene":
                gffs = GTF.gene_iterator(GTF.iterator(options.stdin), strict = False )
                f = lambda x: x[0][0].gene_id
                outf = lambda x: "\n".join( [ "\n".join( [ str(y) for y in xx] ) for xx in x] )
            elif options.remove_duplicates == "transcript":
                gffs = GTF.transcript_iterator(GTF.iterator(options.stdin), strict = False)
                f = lambda x: x[0].transcript_id
                outf = lambda x: "\n".join( [ str(y) for y in x] )
            elif options.remove_duplicates == "coordinates":
                gffs = GTF.chunk_iterator(GTF.iterator(options.stdin) )
                f = lambda x: x[0].contig+"_"+str(x[0].start)+"-"+str(x[0].end)
                outf = lambda x: "\n".join( [ str(y) for y in x] )    

            store = []

            for entry in gffs:
                ninput += 1
                store.append(entry)
                id = f(entry)
                counts[id] += 1

            # Assumes GTF file sorted by contig then start
            last_id = ""
            if options.remove_duplicates == "coordinates":
                for entry in store:
                    id = f(entry)
                    if id == last_id:
                        ndiscarded += 1
                        E.info("discarded duplicates for %s: %i" % (id, counts[id]))
                    else:
                        options.stdout.write( outf(entry) + "\n" )
                        noutput += 1
                    last_id = id
                        
            else:
                for entry in store:
                    id = f(entry)
                    if counts[id] == 1:
                        options.stdout.write( outf(entry) + "\n" )
                        noutput += 1
                    else:
                        ndiscarded += 1
                        E.info("discarded duplicates for %s: %i" % (id, counts[id]))

    elif options.sort:

        for gff in GTF.iterator_sorted( GTF.iterator( options.stdin ), sort_order = options.sort ):
            ninput += 1
            options.stdout.write( "%s\n" % str(gff) )                
            noutput += 1
            nfeatures += 1
            
    elif options.set_gene2transcript:

        for gff in GTF.iterator(options.stdin):

            ninput += 1

            gff.gene_id = gff.transcript_id
            options.stdout.write( "%s\n" % str(gff) )                
            noutput += 1
            nfeatures += 1

    elif options.set_protein2transcript:

        for gff in GTF.iterator(options.stdin):
            ninput += 1
            gff.setAttribute( "protein_id", gff.transcript_id)
            options.stdout.write( "%s\n" % str(gff) )                
            noutput += 1
            nfeatures += 1

    elif options.add_protein_id:

        transcript2protein = IOTools.readMap( open( options.add_protein_id, "r") )

        missing = set()
        for gff in GTF.iterator(options.stdin):
            ninput += 1
            if gff.transcript_id not in transcript2protein:
                if gff.transcript_id not in missing:
                    E.debug( "removing transcript '%s' due to missing protein id" % gff.transcript_id)
                    missing.add( gff.transcript_id)
                ndiscarded += 1
                continue
            
            gff.setAttribute( "protein_id", transcript2protein[gff.transcript_id])
            options.stdout.write( "%s\n" % str(gff) )                
            noutput += 1
            nfeatures += 1
        
        E.info("transcripts removed due to missing protein ids: %i" % len(missing))

    elif options.join_exons:

        for exons in GTF.transcript_iterator( GTF.iterator(options.stdin) ):
            ninput += 1
            strand = Genomics.convertStrand( exons[0].strand )
            contig = exons[0].contig
            transid = exons[0].transcript_id
            geneid = exons[0].gene_id
            biotype = exons[0].source
            all_start, all_end = min( [ x.start for x in exons ] ), max( [x.end for x in exons ] )
            y = GTF.Entry()
            y.contig = contig
            y.source = biotype
            y.feature = "transcript"
            y.start = all_start
            y.end = all_end
            y.strand = strand
            y.transcript_id = transid
            y.gene_id = geneid
            options.stdout.write( "%s\n" % str(y ) )

    elif options.merge_genes:
        # merges overlapping genes
        # 
        gffs = GTF.iterator_sorted_chunks( 
            GTF.flat_gene_iterator(GTF.iterator(options.stdin)),
            sort_by = "contig-strand-start" )
        
        def iterate_chunks( gff_chunks ):

            last = gff_chunks.next()
            to_join = [last]

            for gffs in gff_chunks:
                d = gffs[0].start - last[-1].end

                if gffs[0].contig == last[0].contig and gffs[0].strand == last[0].strand:
                    assert gffs[0].start >= last[0].start, \
                        "input file should be sorted by contig, strand and position: d=%i:\nlast=\n%s\nthis=\n%s\n" % \
                        (d, 
                         "\n".join( [str(x) for x in last] ),
                         "\n".join( [str(x) for x in gffs] ) )

                if gffs[0].contig != last[0].contig or \
                        gffs[0].strand != last[0].strand or \
                        d > 0:
                    yield to_join
                    to_join = []

                last = gffs
                to_join.append( gffs )

            yield to_join
            raise StopIteration
        
        for chunks in iterate_chunks( gffs ):
            ninput += 1
            if len(chunks) > 1:
                gene_id = "merged_%s" % chunks[0][0].gene_id
                transcript_id = "merged_%s" % chunks[0][0].transcript_id
                info = ",".join([ x[0].gene_id for x in chunks ])
            else:
                gene_id = chunks[0][0].gene_id
                transcript_id = chunks[0][0].transcript_id
                info = None

            intervals = []
            for c in chunks: intervals += [ (x.start, x.end) for x in c ]

            intervals = Intervals.combine( intervals )
            # take single strand
            strand = chunks[0][0].strand

            for start, end in intervals:
                y = GTF.Entry()
                y.fromGTF( chunks[0][0], gene_id, transcript_id )
                y.start = start
                y.end = end
                y.strand = strand

                if info: y.addAttribute( "merged", info )
                options.stdout.write( "%s\n" % str(y ) )
                nfeatures += 1

            noutput += 1

    elif options.renumber_genes:
        
        map_old2new = {}
        for gtf in GTF.iterator(options.stdin):
            ninput += 1
            if gtf.gene_id not in map_old2new:
                map_old2new[gtf.gene_id ] = options.renumber_genes % (len(map_old2new) + 1)
            gtf.setAttribute("gene_id", map_old2new[gtf.gene_id ] )
            options.stdout.write( "%s\n" % str(gtf) )
            noutput += 1

    elif options.unset_genes:
        
        map_old2new = {}
        for gtf in GTF.iterator(options.stdin):
            ninput += 1
            key = gtf.transcript_id
            if key not in map_old2new:
                map_old2new[key] = options.unset_genes % (len(map_old2new) + 1)
            gtf.setAttribute( "gene_id",  map_old2new[key] )
            options.stdout.write( "%s\n" % str(gtf) )
            noutput += 1

    elif options.renumber_transcripts:
        
        map_old2new = {}
        for gtf in GTF.iterator(options.stdin):
            ninput += 1
            key = (gtf.gene_id, gtf.transcript_id)
            if key not in map_old2new:
                map_old2new[key] = options.renumber_transcripts % (len(map_old2new) + 1)
            gtf.setAttribute( "transcript_id",  map_old2new[key] )
            options.stdout.write( "%s\n" % str(gtf) )
            noutput += 1

    elif options.transcripts2genes:

        transcripts = set()
        genes = set()
        reset_strand = options.reset_strand
        for gtfs in GTF.iterator_transcripts2genes( GTF.iterator(options.stdin) ):

            ninput += 1
            for gtf in gtfs:
                if reset_strand: gtf.strand = "."
                options.stdout.write( "%s\n" % str(gtf) )                
                transcripts.add( gtf.transcript_id )
                genes.add( gtf.gene_id )
                nfeatures += 1
            noutput += 1

        if options.loglevel >= 1:
            options.stdlog.write("# transcripts2genes: transcripts=%i, genes=%i\n" % \
                                     (len(transcripts), len(genes)))

    elif options.rename:

        map_old2new = IOTools.readMap( open(options.filename_filter, "r") )

        if options.rename == "transcript":
            is_gene_id = False
        elif options.rename == "gene":
            is_gene_id = True
            
        for gff in GTF.iterator( options.stdin ):
            ninput += 1

            if is_gene_id:
                if gff.gene_id in map_old2new:
                    gff.setAttribute("gene_id", map_old2new[gff.gene_id])
                else:
                    E.debug( "removing missing gene_id %s" % gff.gene_id)
                    ndiscarded += 1
                    continue

            else:
                if gff.transcript_id in map_old2new:
                    gff.setAttribute("transcript_id", map_old2new[gff.transcript_id])
                else:
                    E.debug( "removing missing transcript_id %s" % gff.transcript_id)
                    ndiscarded += 1
                    continue

            noutput += 1
            options.stdout.write("%s\n" % str(gff))
                
    elif options.filter:

        keep_genes = set()
        if options.filter == "longest-gene":
            iterator = GTF.flat_gene_iterator( GTF.iterator(options.stdin) )
            coords = []
            gffs = []
            for gff in iterator:
                gff.sort( key = lambda x: x.start )
                coords.append( (gff[0].contig,
                                min([x.start for x in gff]), 
                                max([x.end for x in gff]),
                                gff[0].gene_id ) )
                gffs.append( gff )
            coords.sort()
            
            last_contig = None
            max_end = 0
            longest_gene_id = None
            longest_length = None

            for contig, start, end, gene_id in coords:
                ninput += 1
                if contig != last_contig or start >= max_end:
                    if longest_gene_id: keep_genes.add( longest_gene_id )
                    longest_gene_id = gene_id
                    longest_length = end - start
                    max_end = end
                else:
                    if end-start > longest_length:
                        longest_length, longest_gene_id = end-start, gene_id
                last_contig = contig
                max_end = max(max_end, end)

            keep_genes.add(longest_gene_id)
            invert = options.invert_filter
            for gff in gffs:
                keep = gff[0].gene_id in keep_genes

                if (keep and not invert) or (not keep and invert):
                    noutput += 1
                    for g in gff:
                        nfeatures += 1
                        options.stdout.write("%s\n" % g )
                else:
                    ndiscarded += 1
        elif options.filter in ("longest-transcript", "representative-transcript" ):

            iterator = GTF.gene_iterator( GTF.iterator(options.stdin) )

            def selectLongestTranscript( gene ):
                r = []
                for transcript in gene:
                    transcript.sort( key = lambda x: x.start )
                    length = transcript[-1].end - transcript[0].start
                    r.append( (length, transcript) )
                r.sort()
                return r[-1][1]
            
            def selectRepresentativeTranscript( gene ):
                '''select a representative transcript.

                The representative transcript represent the largest number
                of exons over all transcripts.
                '''
                all_exons = []
                for transcript in gene:
                    all_exons.extend( [ (x.start, x.end) for x in transcript if x.feature == "exon"] )
                exon_counts = {}
                for key, exons in itertools.groupby( all_exons ):
                    exon_counts[key] = len(list(exons))
                transcript_counts = []
                for transcript in gene:
                    count = sum( [ exon_counts[(x.start, x.end)] for x in transcript if x.feature == "exon" ] )
                    transcript_counts.append( (count, transcript) )
                transcript_counts.sort()
                return transcript_counts[-1][1]

            if options.filter == "longest-transcript":
                _select = selectLongestTranscript
            elif options.filter == "representative-transcript":
                _select = selectRepresentativeTranscript

            for gene in iterator:
                ninput += 1
                transcript = _select( gene )
                noutput += 1
                for g in transcript:
                    nfeatures += 1
                    options.stdout.write("%s\n" % g )

        elif options.filter in ("gene", "transcript"):

            if options.filename_filter:
                
                ids, nerrors = IOTools.ReadList( open(options.filename_filter, "r") )
                E.info( "read %i ids" % len(ids))

                ids = set(ids)
                by_gene = options.filter == "gene"
                by_transcript = options.filter == "transcript"
                invert = options.invert_filter

                reset_strand = options.reset_strand
                for gff in GTF.iterator(options.stdin):

                    ninput += 1

                    keep = False
                    if by_gene: keep = gff.gene_id in ids
                    if by_transcript: keep = gff.transcript_id in ids
                    if (invert and keep) or (not invert and not keep): continue

                    if reset_strand: gff.strand = "."

                    options.stdout.write( "%s\n" % str(gff) )                
                    nfeatures += 1
                    noutput += 1

            elif options.sample_size:

                if options.filter == "gene":
                    iterator = GTF.flat_gene_iterator( GTF.iterator(options.stdin) )
                elif options.filter == "transcript":
                    iterator = GTF.transcript_iterator( GTF.iterator(options.stdin) )
                if options.min_exons_length:
                    iterator = GTF.iterator_min_feature_length( iterator, 
                                                                min_length = options.min_exons_length,
                                                                feature = "exon" )

                data = [ x for x in iterator ]
                ninput = len(data)
                if len(data) > options.sample_size:
                    data = random.sample( data, options.sample_size )

                for d in data:
                    noutput += 1
                    for dd in d:
                        nfeatures += 1
                        options.stdout.write( str(dd) + "\n" )

            else:
                assert False, "please supply either a filename with ids to filter with (--apply) or a sample-size."
        
    elif options.exons2introns:

        for gffs in GTF.flat_gene_iterator(GTF.iterator(options.stdin)):

            ninput += 1

            cds_ranges = GTF.asRanges( gffs, "CDS" )
            exon_ranges = GTF.asRanges( gffs, "exon" )
            input_ranges = Intervals.combine( cds_ranges + exon_ranges )
            
            if len(input_ranges) > 1:
                last = input_ranges[0][1]
                output_ranges = []
                for start, end in input_ranges[1:]:
                    output_ranges.append( (last, start) )
                    last = end

                if options.intron_border:
                    b = options.intron_border
                    output_ranges = [ (x[0]+b, x[1]-b ) for x in output_ranges ]

                if options.intron_min_length:
                    l = options.intron_min_length
                    output_ranges = [ x for x in output_ranges if x[1]-x[0] > l ]
            
                for start, end in output_ranges:
                    
                    entry = GTF.Entry()
                    entry.copy( gffs[0] )
                    entry.clearAttributes()
                    entry.transcript_id = "merged"
                    entry.feature = "intron"
                    entry.start = start
                    entry.end = end
                    options.stdout.write( "%s\n" % str( entry ) )
                    nfeatures += 1                    
                noutput += 1
            else:
                ndiscarded += 1

    elif options.set_score2distance:

        for gffs in GTF.transcript_iterator(GTF.iterator(options.stdin)):
            ninput += 1
            strand = Genomics.convertStrand( gffs[0].strand )
            all_start, all_end = min( [ x.start for x in gffs ] ), max( [x.end for x in gffs ] ) 

            if strand != ".":
                t = 0            
                if strand == "-": gffs.reverse()
                for gff in gffs:
                    gff.score = t
                    t += gff.end - gff.start

                if strand == "-": gffs.reverse()
            for gff in gffs:
                options.stdout.write( "%s\n" % str( gff ) )
                nfeatures += 1
            noutput += 1

    elif options.remove_overlapping:
        
        index = GTF.readAndIndex( GFF.iterator( IOTools.openFile( options.remove_overlapping, "r" ) ) )
        
        for gffs in GTF.transcript_iterator(GTF.iterator(options.stdin)):
            ninput += 1
            found = False
            for e in gffs:
                if index.contains( e.contig, e.start, e.end ):
                    found = True
                    break
            
            if found:
                ndiscarded += 1    
            else:
                noutput += 1
                for e in gffs: 
                    nfeatures += 1
                    options.stdout.write( "%s\n" % str(e ) )

    elif options.intersect_transcripts:
        
        for gffs in GTF.gene_iterator(GTF.iterator(options.stdin), strict=options.strict ):
            
            ninput += 1
            r = []
            for g in gffs:
                if options.with_utr:
                    ranges = GTF.asRanges( g, "exon" )
                else:
                    ranges = GTF.asRanges( g, "CDS" )
                r.append( ranges )

            result = r[0]
            for x in r[1:]:
                result = Intervals.intersect( result, x )
            
            entry = GTF.Entry()
            entry.copy( gffs[0][0] )
            entry.clearAttributes()
            entry.transcript_id = "merged"
            entry.feature = "exon"
            for start, end in result:
                entry.start = start
                entry.end = end
                options.stdout.write( "%s\n" % str( entry ) )
                nfeatures += 1
                
            noutput += 1
    else:
        for gffs in GTF.flat_gene_iterator(GTF.iterator(options.stdin), strict=options.strict ):

            ninput += 1

            cds_ranges = GTF.asRanges( gffs, "CDS" )
            exon_ranges = GTF.asRanges( gffs, "exon" )

            # sanity checks
            strands = set( [x.strand for x in gffs ] )
            contigs = set( [x.contig for x in gffs ] )
            if len(strands) > 1:
                raise ValueError( "can not merge gene '%s' on multiple strands: %s" % (gffs[0].gene_id, str(strands)))

            if len(contigs) > 1:
                raise ValueError( "can not merge gene '%s' on multiple contigs: %s" % (gffs[0].gene_id, str(contigs)))

            strand = Genomics.convertStrand( gffs[0].strand )

            if cds_ranges and options.with_utr:
                cds_start, cds_end = cds_ranges[0][0], cds_ranges[-1][1]
                midpoint = ( cds_end - cds_start ) / 2 + cds_start

                utr_ranges = []
                for start, end in Intervals.truncate( exon_ranges, cds_ranges ):
                    if end - start > 3:
                        if strand == ".":
                            feature = "UTR"
                        elif strand == "+":
                            if start < midpoint:
                                feature = "UTR5"
                            else:
                                feature = "UTR3"
                        elif strand == "-":
                            if start < midpoint:
                                feature = "UTR3"
                            else:
                                feature = "UTR5"
                        utr_ranges.append( (feature, start,end) )
                output_feature = "CDS"
                output_ranges = cds_ranges
            else:
                output_feature = "exon"
                output_ranges = exon_ranges
                utr_ranges = []

            result = []

            if options.merge_exons:
                # need to combine per feature - skip
                # utr_ranges = Intervals.combineAtDistance( utr_ranges, options.merge_exons_distance )

                output_ranges = Intervals.combineAtDistance( output_ranges, options.merge_exons_distance )

                for feature, start, end in utr_ranges:
                    entry = GTF.Entry()
                    entry.copy( gffs[0] )
                    entry.clearAttributes()
                    entry.feature = feature
                    entry.transcript_id = "merged"
                    entry.start = start
                    entry.end = end
                    result.append( entry )

                for start, end in output_ranges:

                    entry = GTF.Entry()
                    entry.copy( gffs[0] )
                    entry.clearAttributes()
                    entry.transcript_id = "merged"
                    entry.feature = output_feature
                    entry.start = start
                    entry.end = end
                    result.append( entry )

            elif options.merge_transcripts:

                entry = GTF.Entry()
                entry.copy( gffs[0] )
                entry.clearAttributes()
                entry.transcript_id = entry.gene_id
                entry.start = output_ranges[0][0]
                entry.end = output_ranges[-1][1]
                result.append( entry )

            elif options.merge_introns:

                if len( output_ranges ) >= 2:
                    entry = GTF.Entry()
                    entry.copy( gffs[0] )
                    entry.clearAttributes()
                    entry.transcript_id = entry.gene_id
                    entry.start = output_ranges[0][1]
                    entry.end = output_ranges[-1][0]
                    result.append( entry )
                else:
                    ndiscarded += 1
                    continue

            result.sort( key=lambda x: x.start )

            for x in result:
                options.stdout.write( "%s\n" % str(x) )
                nfeatures += 1
            noutput += 1

    E.info("ninput=%i, noutput=%i, nfeatures=%i, ndiscarded=%i" % (ninput, noutput, nfeatures, ndiscarded) )
    E.Stop()


if __name__ == "__main__":
    sys.exit( main( sys.argv ))



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id: diff_bed.py 2866 2010-03-03 10:18:49Z andreas $
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
"""
diff_bed.py - count differences between several bed files
=========================================================

:Author: Andreas Heger
:Release: $Id: diff_bed.py 2866 2010-03-03 10:18:49Z andreas $
:Date: |today|
:Tags: Python

Purpose
-------

Compute overlap statistics between multiple bed files. For each pairwise
comparison, this script outputs the number of intervals (exons) and
bases overlapping.

Using the ``--update`` option, a table can be incrementally updated with
additional comparisons.

The strand of intervals is ignored in comparisons.

Usage
-----

Type::

   python <script_name>.py --help

for command line help.

Code
----

""" 

import os
import sys
import re
import optparse

import CGAT.Experiment as E
import CGAT.IOTools as IOTools
import CGAT.Bed as Bed
import numpy

class Counter:

    mPercentFormat = "%5.2f"

    def __init__(self):
        pass
        
    def getHeader( self ):
        h = []
        for a in ( "exons", "bases" ):
            for b in ("total", "ovl", "unique"):
                for c in ("1","2"):
                    h.append( "n" + a + "_" + b + c )
        for a in ( "exons", "bases" ):
            for b in ("ovl", "unique"):
                for c in ("1","2"):
                    h.append( "p" + a + "_" + b + c )

        return "\t".join(h)

    @E.cachedmethod
    def buildIndex( self, filename ):
        return Bed.readAndIndex( IOTools.openFile(filename, "r") )

    def _count( self, filename, idx ):
        '''count filename against idx.'''

        overlapping_genes = set()
        genes = set()

        # iterate over exons
        infile = IOTools.openFile( filename, "r" )
        it = Bed.bed_iterator( infile )

        nexons, nexons_overlapping = 0, 0
        nbases, nbases_overlapping = 0, 0
        for this in it:
            nexons += 1
            nbases += this.end - this.start 
            
            try:
                intervals = list(idx[this.contig].find( max(0, this.start), this.end ))
            except KeyError:
                continue
            except Exception, msg:
                raise Exception( "error while processing %s, msg=%s" % (filename, msg ) )
            if len(intervals) == 0: 
                continue

            nexons_overlapping += 1
            start, end = this.start, this.end 
            counts = numpy.zeros( end - start, numpy.int )
            for other_start,other_end,other_value in intervals:
                for x in range( max(start, other_start ) - start, min(end, other_end ) - start ): 
                    counts[x] += 1
            nbases_overlapping += sum( [1 for x in counts if x > 0 ] )

        infile.close()

        return nexons, nexons_overlapping, nbases, nbases_overlapping

    def count( self, filename1, filename2 ):
        """count overlap between two gtf files."""

        E.info( "counting started for %s versus %s" % (filename1, filename2))

        idx2 = self.buildIndex( filename2 )

        (self.mExons1, self.mExonsOverlapping1,
         self.mBases1, self.mBasesOverlapping1 ) = \
         self._count( filename1, idx2 )

        self.mExonsUnique1 = self.mExons1 - self.mExonsOverlapping1
        self.mBasesUnique1 = self.mBases1 - self.mBasesOverlapping1

        idx1 = self.buildIndex( filename1 )

        (self.mExons2, self.mExonsOverlapping2,
         self.mBases2, self.mBasesOverlapping2 ) = \
         self._count( filename2, idx1 ) 

        self.mExonsUnique2 = self.mExons2 - self.mExonsOverlapping2
        self.mBasesUnique2 = self.mBases2 - self.mBasesOverlapping2

    def __str__(self):

        return "\t".join( map( str, (
                        self.mExons1, self.mExons2,
                        self.mExonsOverlapping1, self.mExonsOverlapping2,
                        self.mExonsUnique1, self.mExonsUnique2,
                        self.mBases1, self.mBases2,
                        self.mBasesOverlapping1, self.mBasesOverlapping2,
                        self.mBasesUnique1, self.mBasesUnique2 ) ) ) + "\t" +\
              "\t".join( map( lambda x: IOTools.prettyPercent( *x), (
                    (self.mExonsOverlapping1, self.mExons1), 
                    (self.mExonsOverlapping2, self.mExons2),
                    (self.mExonsUnique1, self.mExons1), 
                    (self.mExonsUnique2, self.mExons2),
                    (self.mBasesOverlapping1, self.mBases1), 
                    (self.mBasesOverlapping2, self.mBases2),
                    (self.mBasesUnique1, self.mBases1), 
                    (self.mBasesUnique2, self.mBases2)  ) ) )

class CounterTracks(Counter):

    def __init__(self, filename ):
        self.mIndices = Bed.readAndIndex( IOTools.openFile(filename, "r"), 
                                          per_track = True )
        
        
    def getTracks( self ):
        return sorted(self.mIndices.keys())

    def _countIndices( self, idx_in, idx ):
        '''count filename against idx.'''

        overlapping_genes = set()
        genes = set()

        # iterate over exons

        nexons, nexons_overlapping = 0, 0
        nbases, nbases_overlapping = 0, 0
        for contig, ix in idx_in.iteritems():

            # note: add a findall function to ncl
            for start, end, value in ix.find(0, 1000000000):
                nexons += 1
                nbases += end - start
            
                try:
                    intervals = list(idx[contig].find( start, end ))
                except KeyError:
                    continue
            
                if len(intervals) == 0: 
                    continue

                nexons_overlapping += 1
                counts = numpy.zeros( end - start, numpy.int )
                for other_start,other_end,other_value in intervals:
                    for x in range( max(start, other_start ) - start, min(end, other_end ) - start ): 
                        counts[x] += 1
                nbases_overlapping += sum( [1 for x in counts if x > 0 ] )

        return nexons, nexons_overlapping, nbases, nbases_overlapping

    def count( self, filename, track ):
        """count overlap between two gtf files."""

        E.info( "counting started for %s versus %s" % (filename, track))

        (self.mExons1, self.mExonsOverlapping1,
         self.mBases1, self.mBasesOverlapping1 ) = \
         self._count( filename, self.mIndices[track] )

        self.mExonsUnique1 = self.mExons1 - self.mExonsOverlapping1
        self.mBasesUnique1 = self.mBases1 - self.mBasesOverlapping1

        idx = self.buildIndex( filename )

        # count index against index
        (self.mExons2, self.mExonsOverlapping2,
         self.mBases2, self.mBasesOverlapping2 ) = \
         self._countIndices( self.mIndices[track], idx )

        self.mExonsUnique2 = self.mExons2 - self.mExonsOverlapping2
        self.mBasesUnique2 = self.mBases2 - self.mBasesOverlapping2

def main( argv = None ):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv: argv = sys.argv

    # setup command line parser
    parser = E.OptionParser( version = "%prog version: $Id: diff_bed.py 2866 2010-03-03 10:18:49Z andreas $", usage = globals()["__doc__"] )

    parser.add_option("-u", "--update", dest="filename_update", type="string",
                      help="if filename is given, previous results will be read from there and only changed sets will be computed [default=%default]." )

    parser.add_option("-p", "--pattern-id", dest="pattern_id", type="string",
                      help="pattern to convert a filename to an id [default=%default]." )

    parser.add_option("-t", "--tracks", dest="tracks", action="store_true",
                      help="compare files against all tracks in the first file [default=%default]" )
    
    parser.set_defaults(
        filename_update = None,
        pattern_id = "(.*).bed",
        tracks = None,
        )

    ## add common options (-h/--help, ...) and parse command line 
    (options, args) = E.Start( parser, argv = argv )

    if len(args) < 2:
        raise ValueError( "at least two arguments required" )

    if options.filename_update:
        infile = IOTools.openFile( options.filename_update, "r")
        previous_results = {}
        for line in infile:
            if line.startswith("#"): continue
            if line.startswith("set1"): continue
            data = line[:-1].split("\t")
            set1,set2 = data[0],data[1]

            if set1 not in previous_results:
                previous_results[set1] = {}
            if set2 not in previous_results:
                previous_results[set2] = {}

            previous_results[set1][set2] = "\t".join( data[2:] )
            rev = [ (data[x+1], data[x]) for x in range( 2, len(data), 2 ) ]
            previous_results[set2][set1] = "\t".join( IOTools.flatten(rev) ) 
    else:
        previous_results = {}

    pattern_id = re.compile(options.pattern_id)

    def getTitle( x ):
        try:
            return pattern_id.search( x ).groups()[0]
        except AttributeError:
            return x

    ncomputed, nupdated = 0, 0

    if options.tracks:
        counter = CounterTracks( args[0] )
        options.stdout.write( "set1\tset2\t%s\n" % counter.getHeader()  )
        for filename in args[1:]:
            title1 = getTitle( filename )
            for title2 in counter.getTracks():
                
                if previous_results:
                    try:
                        prev = previous_results[title1][title2] 
                    except KeyError:
                        pass
                    else:
                        options.stdout.write( "%s\t%s\t%s\n" % ( (title1, title2, prev ) ))
                        nupdated += 1
                        continue

                counter.count( filename, title2 )
                options.stdout.write( "%s\t%s\t%s\n" % ( (title1, title2, str(counter) ) ))
                ncomputed += 1
    else:
        counter = Counter()
        options.stdout.write( "set1\tset2\t%s\n" % counter.getHeader()  )

        for x in range( len(args)):

            title1 = getTitle( args[x] )

            for y in range( 0,x):
                title2 = getTitle( args[y] )
                if previous_results:
                    try:
                        prev = previous_results[title1][title2] 
                    except KeyError:
                        pass
                    else:
                        options.stdout.write( "%s\t%s\t%s\n" % ( (title1, title2, prev ) ))
                        nupdated += 1
                        continue

                counter.count( args[x], args[y] )
                options.stdout.write( "%s\t%s\t%s\n" % ( (title1, title2, str(counter) ) ))
                ncomputed += 1

    E.info( "nupdated=%i, ncomputed=%i" % (nupdated, ncomputed))
    E.Stop()


if __name__ == "__main__":
    sys.exit( main( sys.argv) )



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
mali_extract.py - 
======================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

.. todo::
   
   describe purpose of the script.

Usage
-----

Example::

   python mali_extract.py --help

Type::

   python mali_extract.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import getopt
import math

USAGE="""python %s [OPTIONS] < mali > filtered

Extract sequences from a multiple alignment

Version = $Id: mali_extract.py 2782 2009-09-10 11:40:29Z andreas $

Options:
-h, --help                      print this message.
-v, --verbose=                  loglevel.
-s, --subset=                   subset of ids to select
-c, --components                filename with components to be analyses separately in the multiple alignment
""" % sys.argv[0]

param_long_options=["verbose=", "help", "subset=", "components=", "version" ]

param_short_options="v:ho:s:c::"

param_loglevel = 1

param_gap_char = "-"
param_mask_char = "x"

param_subset = None
param_filename_components = None

import CGAT.Experiment as E
import CGAT.MaliIO as MaliIO
            
##------------------------------------------------------------
if __name__ == '__main__':

    try:
        optlist, args = getopt.getopt(sys.argv[1:], param_short_options, param_long_options)
    except getopt.error, msg:
        print USAGE, msg
        sys.exit(2)

    for o,a in optlist:
        if o in ( "-v", "--verbose" ):
            param_loglevel = int(a)
        elif o in ( "--version", ):
            print "version="
            sys.exit(0)
        elif o in ( "-h", "--help" ):
            print USAGE
            sys.exit(0)
        elif o in ("-s", "--subset"):
            param_subset = a
        elif o == ("-c", "--components"):
            param_filename_components = a

    if param_loglevel >= 1:
        print E.GetHeader()
        print E.GetParams()
            
    ## 1. read multiple alignment in fasta format
    all_mali, all_identifiers = MaliIO.readFasta( sys.stdin )

    if len(all_identifiers) == 0:
        raise "alignment is empty."

    if param_loglevel >= 1:
        print "# read mali with %i entries." % len(all_identifiers)

    if param_filename_components:

        infile = open(param_filename_components, "r")
        components = {}
        for line in infile:
            if line[0] == "#": continue
            if line[0] == ">": continue
            a, b = line[:-1].split("\t")[:2]
            if b not in components:
                components[b] = []
            components[b].append(a)
        
        if param_loglevel >= 1:
            print "# read %i components." % len(components)

    if param_subset:
        components = { 'all' : string.split( param_subset, "," ) }

    for key, identifiers in components.items():
        # 1. remove gaps in multiple alignment
        mali = MaliIO.removeGappedColumns( MaliIO.getSubset( all_mali, identifiers),
                                           param_gap_char )

        for i in identifiers:
            print ">%s\n%s\n" % (i, mali[i] )

    if param_loglevel >= 1:        
        print E.GetFooter()
    
    



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
MatlabTools.py - 
======================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Code
----

'''
import sys, string, re
import numpy


def WriteMatrixOld( matrix, separator = "\t"):
    # this is clumsy

    lines = []
    for x in range(0, matrix.shape[0]):
        lines.append( string.join( map(str, matrix[x,]), separator))
        
    return string.join(lines, "\n" )

def WriteMatrix( matrix, outfile = sys.stdout, separator = "\t", format="%f",
                 row_headers = None, col_headers = None):
    """write matrix to stream.
    """
    if col_headers:
        outfile.write( separator + separator.join( col_headers ) + "\n" )
        
    for x in range(0, matrix.shape[0]):
        if row_headers: outfile.write(row_headers[x] + separator )
        outfile.write( string.join( map(lambda x: format % x, matrix[x,]), separator) + "\n")

def ReadMatrix( file,
                separator = "\t",
                numeric_type = numpy.float,
                take = "all",
                headers = False
                ):
    """read a matrix. There probably is a routine for this in Numpy, which
    I haven't found yet.
    """
    
    lines = filter( lambda x: x[0] != "#", file.readlines())

    row_headers, col_headers = [], []

    if headers:
        col_headers = lines[0][:-1].split("\t")[1:]
        del lines[0]

    num_rows = len(lines)
    
    if take != "all":
        num_cols = len(take)
    else:
        l = len(string.split(lines[0][:-1], "\t"))
        if headers:
            take = range( 1, l)
        else:
            take = range( 0, l)            

    num_cols = len(take)
            
    matrix = numpy.zeros( (num_rows, num_cols), numeric_type )
    
    nrow = 0
    for l in lines:
        data = l[:-1].split("\t")
        if headers: row_headers.append( data[0] )
        
        try:
            data = map( lambda x: float(data[x]), take)
        except ValueError:
            print "error parsing data", data
            raise

        matrix[nrow] = data
        nrow += 1

    return matrix, row_headers, col_headers
        

def ReadSparseMatrix( filename,
                      separator = "\t",
                      numeric_type = numpy.float,
                      is_symmetric = None):
    """read sparse matrix."""

    lines = map(lambda x: string.split(x[:-1], separator)[:3],
                filter(lambda x: x[0] != "#", open(filename,"r").readlines()))

    data = map(lambda x: (string.atoi(x[0]), string.atoi(x[1]), string.atof(x[2])), lines)

    num_rows = max( map(lambda x: x[0], data))
    num_cols = max( map(lambda x: x[1], data))   

    if (is_symmetric):
        num_rows= max( num_rows, num_cols)
        num_cols= max( num_rows, num_cols)

    matrix = numpy.zeros( (num_rows, num_cols), numeric_type )

    for row, col, weight in data:
        matrix[row-1,col-1] = weight

    if is_symmetric:
        for col, row, weight in data:
            matrix[row-1,col-1] = weight
            
    return matrix


def ReadBinarySparseMatrix( filename,
                            separator = "\t",
                            numeric_type = numpy.float,
                            is_symmetric = None):
    """read sparse matrix."""

    lines = map(lambda x: string.split(x[:-1], separator)[:2],
                filter(lambda x: x[0] != "#", open(filename,"r").readlines()))

    data = map(lambda x: (string.atoi(x[0]), string.atoi(x[1])), lines)

    num_rows = max( map(lambda x: x[0], data))
    num_cols = max( map(lambda x: x[1], data))   

    if (is_symmetric):
        num_rows= max( num_rows, num_cols)
        num_cols= max( num_rows, num_cols)
    matrix = numpy.zeros( (num_rows, num_cols), numeric_type )


    for row, col in data:
        matrix[row-1,col-1] = 1

    if is_symmetric:
        for col, row in data:
            matrix[row-1,col-1] = 1
            
    return matrix


def readMatrix( infile,
                format = "full",
                separator = "\t",
                numeric_type = numpy.float,
                take = "all",
                headers = True,
                missing = None,
                ):
    """read a matrix from file ane return a numpy matrix.

    formats accepted are:
    * full
    * sparse
    * phylip
    """
    
    row_headers, col_headers = [], []

    lines = filter( lambda x: x[0] != "#", infile.readlines())

    if len(lines) == 0: raise IOError("no input")

    if format == "full":

        if headers:
            col_headers = lines[0][:-1].split("\t")[1:]
            del lines[0]

        num_rows = len(lines)

        if take != "all":
            num_cols = len(take)
        else:
            l = len(string.split(lines[0][:-1], "\t"))
            if headers:
                take = range( 1, l)
            else:
                take = range( 0, l)            

        num_cols = len(take)

        matrix = numpy.zeros( (num_rows, num_cols), numeric_type )

        nrow = 0
        for l in lines:
            data = l[:-1].split("\t")
            if headers: row_headers.append( data[0] )

            if missing == None:
                try:
                    data = map( lambda x: float(data[x]), take)
                except ValueError, msg:
                    raise ValueError( "error %s: data=%s" % (msg, str( data )))
                except IndexError, msg:
                    raise IndexError( "error %s: data=%s" % (msg, str( data )))

            else:
                d = []
                for x in take:
                    try:
                        d.append( float(data[x]) )
                    except ValueError:
                        d.append( missing )
                    except IndexError, msg:
                        raise IndexError( "error %s: data=%s" % (msg, str( data )))

                data = d

            matrix[nrow] = data

            nrow += 1

    elif format == "phylip":
        ## read in symmetric phylip matrices
        ## note: they can wrap around
        if take != "all":
            raise "phylip matrix does not support take - only full matrices are processed."

        if not headers:
            raise "phylip matrix always has headers."
            
        num_rows = int(lines[0].strip())
        
        num_cols = num_rows

        matrix = numpy.zeros( (num_rows, num_cols), numeric_type )
        take = range( 1, num_rows)

        nrow = 0
        ncol = 0        
        for l in lines[1:]:

            data = re.split("\s+", l[:-1])
            
            if ncol == 0:
                row_headers.append( data[0] )
            
            try:
                data = map( float, data[1:len(data)] )
            except ValueError:
                raise "parsing error in conversion to float in line %s" % l

            for x in range(len(data)):
                matrix[nrow][ncol] = data[x]
                ncol += 1

            ## deal with wrapping around
            if ncol == num_cols:
                ncol = 0
                nrow += 1
        
        col_headers = row_headers
        
    return matrix, row_headers, col_headers
        
def writeMatrix( outfile, matrix,
                 format= "full",
                 separator = "\t",
                 value_format="%f",
                 row_headers = None,
                 col_headers = None):
    """write matrix to stream.
    """

    if format == "full":
        if col_headers:
            outfile.write( separator + separator.join( col_headers ) + "\n" )

        for x in range(0, matrix.shape[0]):
            if row_headers: outfile.write(row_headers[x] + separator )
            outfile.write( string.join( map(lambda x: value_format % x, matrix[x,]), separator) + "\n")
            
    elif format == "phylip":
        if not row_headers:
            raise "phylip output requires row headers."

        nrows = len(row_headers)
        outfile.write( "%i\n" % nrows )

        for x in range(0, nrows):
            outfile.write(row_headers[x] + separator )
            outfile.write( separator.join( [value_format % y for y in matrix[x,]]) + "\n" )
            
    else:
        raise "unknown output format %s" % output_format








################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
CGAT.py - CGAT project specific functions
=========================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

The :mod:`CGAT` module contains various utility functions
for working on CGAT projects.

API
----
'''

from ruffus import *

import os, sys, re, shutil
import Experiment as E

PROJECT_ROOT = '/ifs/projects'

def getProjectId():
    '''cgat specific method: get the (obfuscated) project id
    based on the current working directory.
    '''
    curdir = os.path.abspath(os.getcwd())
    if not curdir.startswith( PROJECT_ROOT ):
        raise ValueError( "method getProjectId no called within %s" % PROJECT_ROOT )
    prefixes = len(PROJECT_ROOT.split("/"))
    # patch for projects in mini directory
    if "/mini/" in curdir: prefixes += 1
    rootdir = "/" + os.path.join( *(curdir.split( "/" )[:prefixes+1]) )
    f = os.path.join( rootdir, "sftp", "web" )
    if not os.path.exists(f):
        raise OSError( "web directory at '%s' does not exist" % f )
    target = os.readlink( f )
    return os.path.basename( target )


def getProjectName():
    '''cgat specific method: get the name of the project 
    based on the current working directory.'''

    curdir = os.path.abspath(os.getcwd())
    if not curdir.startswith( PROJECT_ROOT ):
        raise ValueError( "method getProjectName no called within %s" % PROJECT_ROOT )
    prefixes = len(PROJECT_ROOT.split("/"))
    return curdir.split( "/" )[prefixes]

def publish_report( prefix = "", 
                    patterns = [], 
                    project_id = None,
                    prefix_project = "/ifs/projects",
                    ):
    '''publish report into web directory.

    Links export directory into web directory.

    Copies html pages and fudges links to the pages in the
    export directory.

    If *prefix* is given, the directories will start with prefix.

    *patterns* is an optional list of two-element tuples (<pattern>, replacement_string).
    Each substitutions will be applied on each file ending in .html.

    If *project_id* is not given, it will be looked up. This requires
    that this method is called within a subdirectory of PROJECT_ROOT.

    .. note::
       This function is CGAT specific.

    '''

    if not prefix:
        prefix = PARAMS.get( "report_prefix", "" )

    web_dir = PARAMS["web_dir"]
    if project_id == None: 
        project_id = getProjectId()

    src_export = os.path.abspath( "export" )
    dest_report = prefix + "report"
    dest_export = prefix + "export"

    curdir = os.path.abspath( os.getcwd() )

    def _link( src, dest ):
        '''create links. 

        Only link to existing targets.
        '''
        dest = os.path.abspath( os.path.join( PARAMS["web_dir"], dest ) )
        if os.path.exists( dest ):
            os.remove(dest)
        
        if os.path.exists( src ):
            #IMS: check if base path of dest exists. This allows for prefix to be a 
            #nested path structure e.g. project_id/
            if not os.path.exists(os.path.dirname(os.path.abspath(dest))):
                os.mkdir(os.path.dirname(os.path.abspath(dest)))

            os.symlink( os.path.abspath(src), dest )

    def _copy( src, dest ):
        dest = os.path.abspath( os.path.join( PARAMS["web_dir"], dest ) )
        if os.path.exists( dest ): shutil.rmtree( dest )
        if not os.path.exists(src):
            E.warn( "%s does not exist - skipped" % src )
            return
        shutil.copytree( os.path.abspath(src), dest ) 

    # publish export dir via symlinking
    E.info( "linking export directory in %s" % dest_export )
    _link( src_export, dest_export )

    # publish web pages by copying
    E.info( "publishing web pages in %s" % os.path.abspath( os.path.join( PARAMS["web_dir"], dest_report )))
    _copy( os.path.abspath("report/html"), dest_report )

    # substitute links to export and report
    _patterns = [ (re.compile( src_export ), 
                   "http://www.cgat.org/downloads/%(project_id)s/%(dest_export)s" % locals() ), 
                  (re.compile( '(%s)/report' % os.path.join( prefix_project, getProjectName() ) ),
                   "http://www.cgat.org/downloads/%(project_id)s/%(dest_report)s" % locals() ),
                  (re.compile( '(%s)/_static' % curdir),
                   "http://www.cgat.org/downloads/%(project_id)s/%(dest_report)s/_static" % locals() )]
    
    _patterns.extend( patterns )

    for root, dirs, files in os.walk(os.path.join( web_dir, dest_report)):
        for f in files:
            fn = os.path.join( root, f )
            if fn.endswith(".html" ):
                with open( fn ) as inf:
                    data = inf.read()
                for rx, repl in _patterns:
                    data = rx.sub( repl, data )
                outf = open( fn, "w" )
                outf.write( data )
                outf.close()

    E.info( "report has been published at http://www.cgat.org/downloads/%(project_id)s/%(dest_report)s" % locals())

def getMakefiles( makefiles, source_directory = "", ignore_missing = False):
    """get all makefiles that are included in a set of makefiles.

    Keep the order of inclusion.
    """

    read_makefiles = set()
    output_makefiles = []

    def __getMakefiles( makefiles ):

        new_makefiles = []
        for makefile in makefiles:
            if makefile in read_makefiles: continue
            print makefile
            if os.path.exists( makefile ):
                fn = makefile
            elif os.path.exists( os.path.join( source_directory, makefile) ):
                fn = os.path.join( source_directory, makefile  )
            else:
                if ignore_missing:
                    continue
                else:
                    raise IOError, "could not find %s in %s" % (makefile, source_directory)
                
            output_makefiles.append( fn )
            infile = open(fn, "r")
            
            for line in infile:
                if re.match("include\s+(\S+)", line):
                    fn = re.search("include\s+(\S+)", line).groups()[0]
                    # add explicitely given files
                    if os.path.exists( fn ):
                        new_makefile = fn
                    else:
                        new_makefile = os.path.basename( fn )
                        # remove any path name variables
                        new_makefile = new_makefile[new_makefile.find(")")+1:]
                    if new_makefile not in read_makefiles:
                        new_makefiles.append( new_makefile )
            infile.close()

            read_makefiles.add( makefile )
            
        if new_makefiles:
            __getMakefiles( new_makefiles )

    __getMakefiles( makefiles )
    
    return output_makefiles

def getScripts( makefiles, source_directory ):
    """extract all python and perl scripts from a set of makefiles."""
    
    scripts = set()
    
    for makefile in makefiles:
        for line in open( makefile,"r"):        
            if re.search( "python", line):
                try:
                    python_scripts = re.search( "python\s+(\S+.py)", line ).groups()
                except AttributeError:
                    continue
                
                for s in python_scripts: scripts.add( ("python", s) )
    
            if re.search( "perl", line):
                try:
                    perl_scripts = re.search( "perl\s+(\S+.pl)", line ).groups()
                except AttributeError:
                    continue
                
                for s in perl_scripts: scripts.add( ("perl", s) )
    
    return scripts

def getModules( modules, scriptdirs, libdirs):
    """extract all imported libraries (modules) from a set of (python) scripts.

    Libraries that are not found are ignored and assumed to be
    installed systemwide.
    """

    read_modules = set()
    system_modules = set()
    
    def __getModules( modules ):
        
        new_modules = set()

        for lib in modules:
            if lib in read_modules or lib in system_modules: continue
            for x in scriptdirs + libdirs:
                if os.path.exists( x + lib ):
                    for line in open( x + lib,"r"):
                        if re.match("import\s+(\S+)", line):
                            ll = re.search("import\s+(\S+)", line).groups()[0]
                            ll = filter( lambda x: x != "", map(lambda x: x.strip(), ll.split(",")))
                            for l in ll:
                                new_modules.add( l + ".py" )
                                
                    read_modules.add( lib )                                
                    break
            else:
                system_modules.add( lib )

        if new_modules:
            __getModules( new_modules )

    __getModules( modules )

    read_modules = read_modules.difference( modules )
    
    return read_modules, system_modules




################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
psl2wiggle_stats.py - intersect psl and wiggle files to compute stats
=====================================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

.. todo::
   
   describe purpose of the script.

Usage
-----

Example::

   python psl2wiggle_stats.py --help

Type::

   python psl2wiggle_stats.py --help

for command line help.

Documentation
-------------

Code
----

'''

import sys
import re
import string
import optparse
import time
import os
import glob

import CGAT.Experiment as E
import CGAT.Stats as Stats
import CGAT.Blat as Blat

import CGAT.Wiggle as Wiggle
import alignlib

if __name__ == '__main__':

    parser = E.OptionParser( version = "%prog version: $Id: psl2wiggle_stats.py 2781 2009-09-10 11:33:14Z andreas $", usage = globals()["__doc__"] )

    parser.add_option("-g", "--genome-file", dest="genome_file", type="string",
                      help="filename with genome."  )

    parser.add_option("--wiggle-files", dest="wiggle_files", type="string",
                      help="glob expression for wiggle files [%default]."  )

    parser.add_option("--prefix", dest="prefix", type="string",
                      help="prefix to add to contig names before lookup [%default]."  )

    parser.add_option( "-z", "--from-zipped", dest="from_zipped", action="store_true",
                       help="input is zipped.")

    parser.add_option( "--test", dest="test", type="int",
                       help="test - stop after # rows of parsing [%default]." )

    parser.add_option( "--with-values", dest="with_values", action="store_true",
                       help="output values in last column [%default]." )

    parser.set_defaults( wiggle_files = "*.data.bz2",
                         from_zipped = False,
                         prefix = "",
                         with_values = False,
                         test = None )

    (options, args) = E.Start( parser, add_pipe_options = True )

    # open indexed access to wiggles
    wiggle_files = glob.glob( options.wiggle_files )
    if not wiggle_files:
        raise IOError( "could not find wiggle files with '%s'" % options.wiggle_files )

    index = Wiggle.WiggleMultiIndexedAccess( wiggle_files,
                                             keep_open = True,
                                             use_cache = False )

    iterator = Blat.BlatIterator( sys.stdin )

    ninput, noutput, nskipped = 0, 0, 0

    options.stdout.write( "query\tnali\t%s" % ("\t".join( Stats.DistributionalParameters().getHeaders() )))
    if options.with_values:
        options.stdout.write( "\tvalues" )
    options.stdout.write("\n" )

    while 1:

        if options.test and ninput >= options.test:
            break

        match = iterator.next()
        
        if match == None: break
        
        ninput += 1

        if options.loglevel >= 2:
            options.stdlog.write( str(match) + "\n" )

        # psl always matches on the forward strand

        map_genome2query = alignlib.makeAlignmentBlocks()
        f = alignlib.AlignmentFormatBlat( "%i\t%i\t%i\t%i\t%s\t%s\t%s\n" % (\
                match.mSbjctFrom,
                match.mSbjctTo,
                match.mQueryFrom,
                match.mQueryTo,
                match.mSbjctBlockStarts,
                match.mQueryBlockStarts,
                match.mBlockSizes ) )
        f.copy( map_genome2query )

        data = index.get( options.prefix + match.mSbjctId,
                          match.mSbjctFrom,
                          match.mSbjctTo )

        values = []
        for x, vv in data:
            for v in vv:
                if map_genome2query.mapRowToCol( x ) >= 0:
                    values.append( v )
                x += 1
        if len(values) == 0:
            nskipped += 1
            continue
            
        noutput += 1

        if options.loglevel >= 2:
            options.stdlog.write( "# %s\n" % ",".join( [ "%5.3f" % v for v in values] ))

        s = Stats.DistributionalParameters( values )
        options.stdout.write( "%s\t%i\t%s" % (match.mQueryId,
                                                match.mNMismatches + match.mNMatches,
                                                str( s ) ) )
        
        if options.with_values:
            options.stdout.write( "\t%s" % ",".join( [ "%5.3f" % v for v in values] ))
            
        options.stdout.write("\n" )    

    if options.loglevel >= 1:
        options.stdlog.write( "# ninput=%i, noutput=%i, nskipped=%i\n" % (ninput, noutput, nskipped) )

    E.Stop()



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
csv_rename.py - rename columns in a table
=========================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

rename columns in a csv file

Usage
-----

Example::

   csv_rename.py gene=id < stdin

Type::

   python csv_rename.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import getopt
import time
import optparse
import math
import tempfile

import CGAT.Experiment as E
import csv

if __name__ == "__main__":

    parser = E.OptionParser( version = "%prog version: $Id: csv_rename.py 2782 2009-09-10 11:40:29Z andreas $")

    parser.add_option( "-r", "--remove", dest="remove", action="store_true",
                       help="remove specified columns, keep all others." )

    parser.add_option( "-u", "--unique", dest="unique", action="store_true",
                       help="output rows are uniq." )

    parser.add_option( "-f", "--filename-fields", dest="filename_fields", type="string",
                       help="filename with field information.")

    parser.set_defaults(
        filename_fields = None,
        )

    (options, args) = E.Start( parser,
                                        add_csv_options  = True )
    mapper = {}
    for x in args:
        a,b = x.split("=")
        mapper[a.strip()] = b.strip()
    
    while 1:
        line = sys.stdin.readline()
        
        if not line:
            E.Stop()
            sys.exit(0)
        
        if line[0] == "#": 
            sys.stdout.write( line )
            continue
        
        break

    header = []
    nreplaced = 0
    for x in line[:-1].split():
        if x in mapper:
            nreplaced += 1
            header.append( mapper[x] )
        else:
            header.append( x )
    
    options.stdout.write( "\t".join( header ) + "\n" )
    nlines = 0
    for line in sys.stdin:
        nlines += 1
        sys.stdout.write( line )
            
    if options.loglevel >= 1:
        ninput = len(header)
        noutput = ninput
        options.stdout.write( "# ninput=%i, noutput=%i, nreplaced=%i, nlines=%i\n" % (ninput,noutput,nreplaced,nlines) )

    E.Stop()
